from __future__ import division, print_function

import cmath
import time
from copy import copy
import os
import argparse
from collections import OrderedDict
from timeit import default_timer as timer
from typing import Tuple
from inspect import getfullargspec

import numpy as np
from numpy import pi, radians, sin, cos, sqrt, clip
from numpy.random import poisson, uniform, randn, rand
from numpy.polynomial.legendre import leggauss
from scipy.integrate import simps
from scipy.special import j1 as J1
from scipy.special import gamma

try:
    from numba import njit, prange
    # SAS_NUMBA: 0=None, 1=CPU, 2=GPU
    SAS_NUMBA = int(os.environ.get("SAS_NUMBA", "1"))
    USE_NUMBA = SAS_NUMBA > 0
    USE_CUDA = SAS_NUMBA > 1
except ImportError:
    # Identity decorator @njit or @njit(...)
    njit = lambda f, *args, **kw: f if callable(f) else (lambda k: k)
    USE_NUMBA = USE_CUDA = False

# Definition of rotation matrices comes from wikipedia:
#    https://en.wikipedia.org/wiki/Rotation_matrix#Basic_rotations
def Rx(angle):
    """Construct a matrix to rotate points about *x* by *angle* degrees."""
    a = radians(angle)
    R = [[1, 0, 0],
         [0, +cos(a), -sin(a)],
         [0, +sin(a), +cos(a)]]
    return np.array(R)

def Ry(angle):
    """Construct a matrix to rotate points about *y* by *angle* degrees."""
    a = radians(angle)
    R = [[+cos(a), 0, +sin(a)],
         [0, 1, 0],
         [-sin(a), 0, +cos(a)]]
    return np.array(R)

def Rz(angle):
    """Construct a matrix to rotate points about *z* by *angle* degrees."""
    a = radians(angle)
    R = [[+cos(a), -sin(a), 0],
         [+sin(a), +cos(a), 0],
         [0, 0, 1]]
    return np.array(R)

def pol2rec(r, theta, phi):
    """
    Convert from 3D polar coordinates to rectangular coordinates.
    """
    theta, phi = radians(theta), radians(phi)
    x = +r * sin(theta) * cos(phi)
    y = +r * sin(theta) * sin(phi)
    z = +r * cos(theta)
    return x, y, z

def jitter(theta, phi, psi):
    r"""
    Return the jitter transform to rotate a set of points.
    View is in degrees using nautical angles with roll $\psi$ around $c$,
    pitch $\theta$ around $b$ and yaw $\phi$ around $a$.

    **unused**
    """
    return Rx(phi) @ Ry(theta) @ Rz(psi)

def rotation(theta, phi, psi):
    r"""
    Return a rotation matrix to apply to a set of points.
    View is in degrees using a $z$-$y$-$z$ rotation sequence of Euler angles
    $\phi$-$\theta$-$\psi$.  The $c$-axis of the shape starts along $z$ and
    the $b$-axis starts along $y$.
    """
    return Rz(phi) @ Ry(theta) @ Rz(psi)

def invert_view(qx, qy, view):
    r"""
    Return $(q_a, q_b, q_c)$ for the $(\theta, \phi, \psi)$ view angle at
    detector pixel corresponding to $(q_x, q_y)$.  View is in degrees using
    a $z$-$y$-$z$ sequence of Euler angles $\phi$-$\theta$-$\psi$.
    """
    theta, phi, psi = view
    Rinv = Rz(-psi) @ Ry(-theta) @ Rz(-phi)
    q = np.vstack((qx.flatten(), qy.flatten(), 0*qx.flatten()))
    return Rinv @ q

def apply_view(points, view):
    r"""
    Return $(p_x, p_y, p_x)$ rotated by the $(\theta, \phi, \psi)$ view angle.
    View is in degrees using a $z$-$y$-$z$ sequence of Euler angles
    $\phi$-$\theta$-$\psi$.
    """
    R = rotation(*view)
    return points @ R.T

class Shape:
    rotation = np.eye(3)
    center = np.array([0., 0., 0.])[:, None]
    r_max = None
    is_magnetic = False

    def volume(self):
        # type: () -> float
        raise NotImplementedError()

    def sample(self, density):
        # type: (float) -> Tuple[np.ndarray, np.ndarray]
        """
        Returns arrays (rho[N], points[N, 3]).
        """
        raise NotImplementedError()

    def dims(self):
        # type: () -> Tuple[float, float, float]
        raise NotImplementedError()

    def rotate(self, theta, phi, psi):
        """See :func:`rotation` for details on the rotation matrix."""
        self.rotation = rotation(theta, phi, psi) @ self.rotation
        return self

    def shift(self, x, y, z):
        self.center = self.center + np.array([x, y, z])[:, None]
        return self

    def _adjust(self, points):
        points = self.rotation @ points.T + self.center
        return points.T

    def r_bins(self, q, over_sampling=1, r_step=None):
        return r_bins(q, r_max=self.r_max, r_step=r_step,
                      over_sampling=over_sampling)

class Composite(Shape):
    def __init__(self, shapes, center=(0, 0, 0), orientation=(0, 0, 0)):
        self.shapes = shapes
        self.rotate(*orientation)
        self.shift(*center)

        # Find the worst case distance between any two points amongst a set
        # of shapes independent of orientation.  This could easily be a
        # factor of two worse than necessary, e.g., a pair of thin rods
        # end-to-end vs the same pair side-by-side.
        distances = [((s1.r_max + s2.r_max)/2
                      + sqrt(np.sum((s1.center - s2.center)**2)))
                     for s1 in shapes
                     for s2 in shapes]
        self.r_max = max(distances + [s.r_max for s in shapes])
        self.volume = sum(shape.volume for shape in self.shapes)

    def sample(self, density):
        values, points = zip(*(shape.sample(density) for shape in self.shapes))
        return np.hstack(values), self._adjust(np.vstack(points))

class Box(Shape):
    def __init__(self, a, b, c,
                 value, center=(0, 0, 0), orientation=(0, 0, 0)):
        self.value = np.asarray(value)
        self.rotate(*orientation)
        self.shift(*center)
        self.a, self.b, self.c = a, b, c
        self._scale = np.array([a/2, b/2, c/2])[None, :]
        self.r_max = sqrt(a**2 + b**2 + c**2)
        self.dims = a, b, c
        self.volume = a*b*c

    def sample(self, density):
        num_points = poisson(density*self.volume)
        points = self._scale*uniform(-1, 1, size=(num_points, 3))
        values = self.value.repeat(points.shape[0])
        return values, self._adjust(points)

class Superball(Shape):
    def __init__(self, a, p,
                 value, center=(0, 0, 0), orientation=(0, 0, 0)):
        self.value = np.asarray(value)
        self.rotate(*orientation)
        self.shift(*center)
        self.a, self.p = a, p
        self._scale = a/2
        # Solve for rounded corner radius x = y = z:
        #    x^2p + y^2p + z^2p = 3 x^2p = (a/2)^2p
        #    => x = a / 2 root[2p](3)
        #    => d = 2r = 2x root(3) = root(3)/root[2p](3) a = 3^(p-1)/2p a
        #self.r_max = 3**((p-1)/(2*p)) * a  # Too short---don't know why.
        self.r_max = sqrt(3)*a
        self.dims = a, a, a
        g1 = gamma(1.0 / (2.0 * p))
        g3 = gamma(3.0 / (2.0 * p))
        self.volume = a**3 / 12.0 / p**2 * g1**3 / g3

    def sample(self, density):
        # Sample from cube[-a/2, a/2]
        num_points = poisson(density*self.a**3)
        points = uniform(-1, 1, size=(num_points, 3))
        # Trim points outside maximum "squared radius", x^2p + y^2p + z^2p < 1 
        radius_sq = np.sum((points**2)**self.p, axis=1)
        points = points[radius_sq <= 1]
        values = self.value.repeat(points.shape[0])
        return values, self._adjust(self._scale*points)

class EllipticalCylinder(Shape):
    def __init__(self, ra, rb, length,
                 value, center=(0, 0, 0), orientation=(0, 0, 0)):
        self.value = np.asarray(value)
        self.rotate(*orientation)
        self.shift(*center)
        self.ra, self.rb, self.length = ra, rb, length
        self._scale = np.array([ra, rb, length/2])[None, :]
        self.r_max = sqrt(4*max(ra, rb)**2 + length**2)
        self.dims = 2*ra, 2*rb, length
        self.volume = pi*ra*rb*length

    def sample(self, density):
        # randomly sample from a box of side length 2*r, excluding anything
        # not in the cylinder
        num_points = poisson(density*4*self.ra*self.rb*self.length)
        points = uniform(-1, 1, size=(num_points, 3))
        radius_sq = points[:, 0]**2 + points[:, 1]**2
        points = points[radius_sq <= 1]
        values = self.value.repeat(points.shape[0])
        return values, self._adjust(self._scale*points)

class EllipticalBicelle(Shape):
    def __init__(self, ra, rb, length,
                 thick_rim, thick_face,
                 value_core, value_rim, value_face,
                 center=(0, 0, 0), orientation=(0, 0, 0)):
        self.rotate(*orientation)
        self.shift(*center)
        self.value = value_core
        self.ra, self.rb, self.length = ra, rb, length
        self.thick_rim, self.thick_face = thick_rim, thick_face
        self.value_rim, self.value_face = value_rim, value_face

        # reset cylinder to outer dimensions for calculating scale, etc.
        ra = self.ra + self.thick_rim
        rb = self.rb + self.thick_rim
        length = self.length + 2*self.thick_face
        self._scale = np.array([ra, rb, length/2])[None, :]
        self.r_max = sqrt(4*max(ra, rb)**2 + length**2)
        self.dims = 2*ra, 2*rb, length
        self.volume = pi*ra*rb*length

    def sample(self, density):
        # randomly sample from a box of side length 2*r, excluding anything
        # not in the cylinder
        ra = self.ra + self.thick_rim
        rb = self.rb + self.thick_rim
        length = self.length + 2*self.thick_face
        num_points = poisson(density*4*ra*rb*length)
        points = uniform(-1, 1, size=(num_points, 3))
        radius = points[:, 0]**2 + points[:, 1]**2
        points = points[radius <= 1]
        # set all to core value first
        values = np.full_like(points[:, 0], self.value)
        # then set value to face value if |z| > face/(length/2))
        values[abs(points[:, 2]) > self.length/(self.length + 2*self.thick_face)] = self.value_face
        # finally set value to rim value if outside the core ellipse
        radius = (points[:, 0]**2*(1 + self.thick_rim/self.ra)**2
                  + points[:, 1]**2*(1 + self.thick_rim/self.rb)**2)
        values[radius>1] = self.value_rim
        return values, self._adjust(self._scale*points)

class TruncatedSphere(Shape):
    """
    Sphere of radius r, with points z < -h truncated.
    """
    def __init__(self, r, h, value, center=(0, 0, 0), orientation=(0, 0, 0)):
        self.value = np.asarray(value)
        self.rotate(*orientation)
        self.shift(*center)
        self.r, self.h = r, h
        # Max distance between points in the shape is the maximum diameter
        self.r_max = 2*r if h >= 0 else 2*sqrt(r**2 - h**2)
        self.dims = self.r_max, self.r_max, r+h
        self.volume = pi*(2*r**3/3 + r**2*h - h**3/3)
        #Vp = pi*(2*r**3/3 + r**2*h - h**3/3)
        #Vm = pi*(2*r**3/3 - r**2*h + h**3/3)
        #Vd = Vp + Vm - 4*pi*r**3/3

    def sample(self, density):
        num_points = poisson(density*np.prod(self.dims))
        points = uniform(-1, 1, size=(num_points, 3))
        # Translate U ~ [-1, 1] in x,y to [-r_trunc/r, r_trunc/r] when
        # truncation starts above the equator, otherwise leave it at [-1, 1].
        # This makes for more efficient sampling since we don't have to
        # consider the maximum diameter.  We already calculated r_max as
        # 2*r_trunc in this case, so just use the ratio of r_max to 2*r.
        points[:, 0:2] *= 0.5*self.r_max/self.r
        # Translate U ~ [-1, 1] in z to [-h/r, 1], with h representing
        # distance below equator.  So:
        #    (U + 1)/2 => [0, 1]
        #    [0, 1] * (1+h/r) => [0, 1+h/r]
        #    [0, 1+h/r] - h/r => [-h/r, 1]
        # Combining:
        #    (U + 1)/2 * (1+h/r) - h/r
        #       = U*(1+h/r)/2 + (1+h/r)/2 - h/r
        #       = U*(1/2 + h/2r) + 1/2 + h/2r - 2h/2r
        ratio = 0.5*self.h/self.r
        points[:, 2] *= (0.5 + ratio)
        points[:, 2] += (0.5 - ratio)
        radius = np.sum(points**2, axis=1)
        points = self.r*points[radius<=1]
        values = self.value.repeat(points.shape[0])
        return values, self._adjust(points)

class TriaxialEllipsoid(Shape):
    def __init__(self, ra, rb, rc,
                 value, center=(0, 0, 0), orientation=(0, 0, 0),
                 magnetism=None):
        self.is_magnetic = (magnetism is not None)
        self.value = np.asarray(value)
        self.magnetism = magnetism if self.is_magnetic else (0., 0., 0.)
        self.rotate(*orientation)
        self.shift(*center)
        self.ra, self.rb, self.rc = ra, rb, rc
        self._scale = np.array([ra, rb, rc])[None, :]
        self.r_max = 2*max(ra, rb, rc)
        self.dims = 2*ra, 2*rb, 2*rc
        self.volume = 4*pi/3 * ra * rb * rc

    def sample(self, density):
        # randomly sample from a box of side length 2*r, excluding anything
        # not in the ellipsoid
        num_points = poisson(density*8*self.ra*self.rb*self.rc)
        points = uniform(-1, 1, size=(num_points, 3))
        radius_sq = np.sum(points**2, axis=1)
        points = self._scale*points[radius_sq <= 1]
        values = self.value.repeat(points.shape[0])
        return values, self._adjust(points)

    def sample_magnetic(self, density):
        values, points = self.sample(density)
        magnetism = np.tile(self.magnetism, (points.shape[0], 1)).T
        return values, magnetism, points

class Helix(Shape):
    def __init__(self, helix_radius, helix_pitch, tube_radius, tube_length,
                 value, center=(0, 0, 0), orientation=(0, 0, 0)):
        self.value = np.asarray(value)
        self.rotate(*orientation)
        self.shift(*center)
        helix_length = helix_pitch * tube_length/sqrt(helix_radius**2 + helix_pitch**2)
        total_radius = self.helix_radius + self.tube_radius
        self.helix_radius, self.helix_pitch = helix_radius, helix_pitch
        self.tube_radius, self.tube_length = tube_radius, tube_length
        self.r_max = sqrt(4*total_radius + (helix_length + 2*tube_radius)**2)
        self.dims = 2*total_radius, 2*total_radius, helix_length
        # small tube radius approximation; for larger tubes need to account
        # for the fact that the inner length is much shorter than the outer
        # length
        self.volume = pi*self.tube_radius**2*self.tube_length

    def points(self, density):
        num_points = poisson(density*4*self.tube_radius**2*self.tube_length)
        points = uniform(-1, 1, size=(num_points, 3))
        radius_sq = points[:, 0]**2 + points[:, 1]**2
        points = points[radius_sq <= 1]

        # Based on math stackexchange answer by Jyrki Lahtonen
        #     https://math.stackexchange.com/a/461637
        # with helix along z rather than x [so tuples in answer are (z, x, y)]
        # and with random points in the cross section (p1, p2) rather than
        # uniform points on the surface (cos u, sin u).
        a, R = self.tube_radius, self.helix_radius
        h = self.helix_pitch
        scale = 1/sqrt(R**2 + h**2)
        t = points[:, 3] * (self.tube_length * scale/2)
        cos_t, sin_t = cos(t), sin(t)

        # rx = R*cos_t
        # ry = R*sin_t
        # rz = h*t
        # nx = -a * cos_t * points[:, 1]
        # ny = -a * sin_t * points[:, 1]
        # nz = 0
        # bx = (a * h/scale) * sin_t * points[:, 2]
        # by = (-a * h/scale) * cos_t * points[:, 2]
        # bz = a*R/scale
        # x = rx + nx + bx
        # y = ry + ny + by
        # z = rz + nz + bz
        u, v = (R - a*points[:, 1]), (a * h/scale)*points[:, 2]
        x = u * cos_t + v * sin_t
        y = u * sin_t - v * cos_t
        z = a*R/scale + h * t

        points = np.hstack((x, y, z))
        values = self.value.repeat(points.shape[0])
        return values, self._adjust(points)

def csbox(a=10, b=20, c=30, da=1, db=2, dc=3, slda=1, sldb=2, sldc=3, sld_core=4):
    core = Box(a, b, c, sld_core)
    side_a = Box(da, b, c, slda, center=((a+da)/2, 0, 0))
    side_b = Box(a, db, c, sldb, center=(0, (b+db)/2, 0))
    side_c = Box(a, b, dc, sldc, center=(0, 0, (c+dc)/2))
    side_a2 = copy(side_a).shift(-a-da, 0, 0)
    side_b2 = copy(side_b).shift(0, -b-db, 0)
    side_c2 = copy(side_c).shift(0, 0, -c-dc)
    shape = Composite((core, side_a, side_b, side_c, side_a2, side_b2, side_c2))
    shape.dims = 2*da+a, 2*db+b, 2*dc+c
    return shape

def barbell(r=20, rbell=50, length=20, rho=2):
    h = sqrt(rbell**2 - r**2)
    top = TruncatedSphere(rbell, h, value=rho, center=(0, 0, length/2+h))
    rod = EllipticalCylinder(r, r, length, value=rho)
    bottom = TruncatedSphere(rbell, h, value=rho, center=(0, 0, -length/2-h),
                             orientation=(180, 0, 0))
    shape = Composite((top, rod, bottom))
    shape.dims = 2*rbell, 2*rbell, length+2*(rbell+h)
    # r_max should be total length?
    shape.r_max = (length + 2*(rbell + h))
    return shape

def capped_cylinder(r=20, rcap=50, length=20, rho=2):
    h = -sqrt(rcap**2 - r**2)
    top = TruncatedSphere(rcap, h, value=rho, center=(0, 0, length/2+h))
    rod = EllipticalCylinder(r, r, length, value=rho)
    bottom = TruncatedSphere(rcap, h, value=rho, center=(0, 0, -length/2-h),
                             orientation=(180, 0, 0))
    shape = Composite((top, rod, bottom))
    shape.dims = 2*r, 2*r, length+2*(rcap+h)
    # r_max is the length of the diagonal + height of the cap for safety.
    # This is a bit larger than necessary, but that's better than truncation.
    shape.r_max = sqrt(length**2 + 4*r**2) + (rcap + h)
    return shape

def _Iqabc(weight, x, y, z, qa, qb, qc):
    """I(q) = |sum V(r) rho(r) e^(1j q.r)|^2 / sum V(r)"""
    #print("calling python")
    Iq = [abs(np.sum(weight*np.exp(1j*(qa_k*x + qb_k*y + qc_k*z))))**2
          for qa_k, qb_k, qc_k in zip(qa.flat, qb.flat, qc.flat)]
    return np.asarray(Iq)
_Iqabcf = _Iqabc

if USE_NUMBA:
    # Override simple numpy solution with numba if available
    def _Iqabc_py(weight, x, y, z, qa, qb, qc):
        #print("calling numba")
        Iq = np.empty_like(qa)
        for j in prange(len(Iq)):
            #total = 0. + 0j
            #for k in range(len(weight)):
            #    total += weight[k]*np.exp(1j*(qa[j]*x[k] + qb[j]*y[k] + qc[j]*z[k]))
            total = np.sum(weight * np.exp(1j*(qa[j]*x + qb[j]*y + qc[j]*z)))
            Iq[j] = abs(total)**2
        return Iq
    sig = "f8[:](f8[:],f8[:],f8[:],f8[:],f8[:],f8[:],f8[:])"
    _Iqabc = njit(sig, parallel=True, fastmath=True)(_Iqabc_py)
    _Iqabcf = njit(sig.replace("f8", "f4"), parallel=True, fastmath=True)(_Iqabc_py)

if USE_CUDA:
    # delayed loading of cuda
    _IQABC_CUDA_KERNELS = {}
    def _get_Iqabc_kernel(dtype):
        #print("calling cuda")
        if not _IQABC_CUDA_KERNELS:
            from numba import cuda
            if not cuda.list_devices():
                raise RuntimeError("no cuda devices found")
            def _kernel_py(weight, x, y, z, qa, qb, qc, Iq):
                j = cuda.grid(1)
                if j < qa.size:
                    total = 0. + 0j
                    for k in range(x.size):
                        total += weight[k]*cmath.exp(1j*(qa[j]*x[k] + qb[j]*y[k] + qc[j]*z[k]))
                    Iq[j] = abs(total)**2
            sig_d = "void(f8[:],f8[:],f8[:],f8[:],f8[:],f8[:],f8[:],f8[:])"
            sig_f = sig_d.replace("f8", "f4")
            kernel_f = cuda.jit(sig_f, parallel=True, fastmath=True)(_kernel_py)
            kernel_d = cuda.jit(sig_d, parallel=True, fastmath=True)(_kernel_py)
            _IQABC_CUDA_KERNELS['f'] = kernel_f
            _IQABC_CUDA_KERNELS['d'] = kernel_d
        kernel = _IQABC_CUDA_KERNELS[dtype.char]
        return kernel

    def _Iqabc(weight, x, y, z, qa, qb, qc):
        Iq = np.empty_like(qa)
        # Apparently numba deals with all the necessary padding of vectors to nice boundaries
        # before transfering to the GPU, so we don't need to do so by hand here.
        threadsperblock = 32
        blockspergrid = (Iq.size + (threadsperblock - 1)) // threadsperblock
        kernel = _get_Iqabc_kernel(qa.dtype)
        kernel[blockspergrid, threadsperblock](weight, x, y, z, qa, qb, qc, Iq)
        return Iq
    _Iqabcf = _Iqabc


if 0 and USE_CUDA:
    ### *** DEPRECATED ***
    ### Variant on the kernel with padding of vectors that is no faster and doesn't appear
    ### to be more correct.  Leave it around for now in case we decide we don't trust numba.
    _IQABC_CUDA_KERNELS = {}
    def _get_Iqabc_kernel(dtype):
        #print("calling cuda")
        if not _IQABC_CUDA_KERNELS:
            from numba import cuda
            if not cuda.list_devices():
                raise RuntimeError("no cuda devices found")
            def _kernel_py(nx, nq, weight, x, y, z, qa, qb, qc, Iq):
                j = cuda.grid(1)
                if j < nq:
                    total = 0. + 0j
                    for k in range(nx):
                        total += weight[k]*cmath.exp(1j*(qa[j]*x[k] + qb[j]*y[k] + qc[j]*z[k]))
                    Iq[j] = abs(total)**2
            sig_d = "void(i4,i4,f8[:],f8[:],f8[:],f8[:],f8[:],f8[:],f8[:],f8[:])"
            sig_f = sig_d.replace("f8", "f4")
            kernel_f = cuda.jit(sig_f, parallel=True, fastmath=True)(_kernel_py)
            kernel_d = cuda.jit(sig_d, parallel=True, fastmath=True)(_kernel_py)
            _IQABC_CUDA_KERNELS['f'] = kernel_f
            _IQABC_CUDA_KERNELS['d'] = kernel_d
        kernel = _IQABC_CUDA_KERNELS[dtype.char]
        return kernel

    def _Iqabc(weight, x, y, z, qa, qb, qc):
        kernel = _get_Iqabc_kernel(qa.dtype)
        nx, nq = len(x), len(qa)
        threadsperblock = 32
        blockspergrid = (nq + (threadsperblock - 1)) // threadsperblock
        weight, x, y, z = pad_vectors(4, weight, x, y, z)
        qa, qb, qc = pad_vectors(threadsperblock, qa, qb, qc)
        Iq = np.empty_like(qa)
        kernel[blockspergrid, threadsperblock](nx, nq, weight, x, y, z, qa, qb, qc, Iq)
        return Iq[:nq]
    _Iqabcf = _Iqabc

    def pad_vectors(boundary, *vectors):
        """
        Yields a list of vectors padded with NaN to a multiple of *boundary*.
        Yields the original vector if the size is already a mulitple of *boundary*.
        """
        for old in vectors:
            old_size = len(old)
            new_size = ((old_size + boundary-1)//boundary)*boundary
            if new_size > old_size:
                new = np.empty(new_size, dtype=old.dtype)
                new[:old_size] = old
                new[old_size:] = np.nan
                yield new
            else:
                yield old

def calc_Iqxy(qx, qy, rho, points, volume=1.0, view=(0, 0, 0), dtype='f'):
    """
    *qx*, *qy* correspond to the detector pixels at which to calculate the
    scattering, relative to the beam along the negative z axis.
    *points* are three columns (x, y, z), one for each sample in the shape.
    *rho* (1e-6/Ang) is the scattering length density of each point.
    *volume* should be 1/number_density.  That is, each of n particles in the
    total value represents volume/n contribution to the scattering.
    *view* rotates the points about the axes using Euler angles for pitch
    yaw and roll for a beam travelling along the negative z axis.
    *dtype* is the numerical precision of the calculation.
    """
    # TODO: maybe slightly faster to rotate points, and drop qc*z
    qx, qy = np.broadcast_arrays(qx, qy)
    qa, qb, qc = invert_view(qx, qy, view)
    rho, volume = np.broadcast_arrays(rho, volume)
    weight = rho*volume
    x, y, z = points.T

    # I(q) = |sum V(r) rho(r) e^(1j q.r)|^2 / sum V(r)
    if np.dtype(dtype) == np.float64:
        weight, x, y, z, qa, qb, qc = [np.asarray(v, 'd') for v in (weight, x, y, z, qa, qb, qc)]
        Iq = _Iqabc(weight, x, y, z, qa.flatten(), qb.flatten(), qc.flatten())
    else:  # float32
        weight, x, y, z, qa, qb, qc = [np.asarray(v, 'f') for v in (weight, x, y, z, qa, qb, qc)]
        Iq = _Iqabcf(weight, x, y, z, qa.flatten(), qb.flatten(), qc.flatten())
    # The scale factor 1e-4 is due to the conversion from rho = 1e-6 squared
    # times the conversion of 1e-8 from inverse angstroms to inverse cm.
    return np.asarray(Iq).reshape(qx.shape) * (1e-4 / np.sum(volume))

def spin_weights(in_spin, out_spin):
    """
    Compute spin cross sections given in_spin and out_spin
    To convert spin cross sections to sld b:
        uu * (sld - m_sigma_x);
        dd * (sld + m_sigma_x);
        ud * (m_sigma_y - 1j*m_sigma_z);
        du * (m_sigma_y + 1j*m_sigma_z);
    weights for spin crosssections: dd du real, ud real, uu, du imag, ud imag
    """

    in_spin = clip(in_spin, 0.0, 1.0)
    out_spin = clip(out_spin, 0.0, 1.0)
    # Previous version of this function took the square root of the weights,
    # under the assumption that
    #
    #     w*I(q, rho1, rho2, ...) = I(q, sqrt(w)*rho1, sqrt(w)*rho2, ...)
    #
    # However, since the weights are applied to the final intensity and
    # are not interned inside the I(q) function, we want the full
    # weight and not the square root.  Anyway no function will ever use
    # set_spin_weights as part of calculating an amplitude, as the weights are
    # related to polarisation efficiency of the instrument. The weights serve to
    # construct various magnet scattering cross sections, which are linear combinations
    # of the spin-resolved cross sections. The polarisation efficiency e_in and e_out
    # are parameters ranging from 0.5 (unpolarised) beam to 1 (perfect optics).
    # For in_spin or out_spin <0.5 one assumes a CS, where the spin is reversed/flipped
    # with respect to the initial supermirror polariser. The actual polarisation efficiency
    # in this case is however e_in/out = 1-in/out_spin.

    norm = 1 - out_spin if out_spin < 0.5 else out_spin

    # The norm is needed to make sure that the scattering cross sections are
    # correctly weighted, such that the sum of spin-resolved measurements adds up to
    # the unpolarised or half-polarised scattering cross section. No intensity weighting
    # needed on the incoming polariser side (assuming that a user), has normalised
    # to the incoming flux with polariser in for SANSPOl and unpolarised beam, respectively.

    weight = (
        (1.0 - in_spin) * (1.0 - out_spin) / norm, # dd
        (1.0 - in_spin) * out_spin / norm,       # du
        in_spin * (1.0 - out_spin) / norm,       # ud
        in_spin * out_spin / norm,             # uu
    )
    return weight

def orth(A, b_hat): # A = 3 x n, and b_hat unit vector
    #return A - np.sum(A*b_hat[:, None], axis=0)[None, :]*b_hat[:, None]
    return A - np.outer(b_hat, b_hat)@A


def magnetic_sld(qx, qy, up_theta, up_phi, rho, rho_m):
    """
    Compute the complex sld for the magnetic spin states.
    Returns effective rho for spin states [dd, du, ud, uu].
    """
    # For q=0 one would see the demagnetising field of the sample, equivalent
    # to direction q_hat = [sqrt(1/2), sqrt(1/2), 0] for a disc shaped sample
    # that is very thin along the beam.
    # Note: This is different from kernel_iq.c, which sets I(0, 0) to zero.
    q_norm = sqrt(qx**2 + qy**2)
    if abs(q_norm) < 1.e-16:
        q_hat = np.array([1., 1., 0.]) / np.sqrt(2)
    else:
        q_hat = np.array([qx, qy, 0]) / q_norm
    M_perp = orth(rho_m, q_hat)  # M = rho_m

    # perpy_hat and perpz_hat are unit vectors spanning up the plane
    # perpendicular to polarisation for SF scattering (avoiding
    # repetitive computation of orthogonal vectors)
    cos_theta, sin_theta = cos(radians(up_theta)), sin(radians(up_theta))
    cos_phi, sin_phi = cos(radians(up_phi)), sin(radians(up_phi))
    p_hat = np.array([sin_theta * cos_phi, sin_theta * sin_phi, cos_theta])
    perpy_hat = np.array([-sin_phi, cos_phi, 0])
    perpz_hat = np.array([-cos_theta * cos_phi, -cos_theta * sin_phi, sin_theta])

    perpx = p_hat @ M_perp
    perpy = perpy_hat @ M_perp
    perpz = perpz_hat @ M_perp

    return (
        rho - perpx,   # dd => sld - D M_perpx
        perpy - 1j * perpz, # du => -D (M_perpy + j M_perpz)
        perpy + 1j * perpz, # ud => -D (M_perpy - j M_perpz)
        rho + perpx,   # uu => sld + D M_perpx
    )

# TODO: provide numba and cuda version
def calc_Iq_magnetic(qx, qy, rho, rho_m, points, volume=1.0, view=(0, 0, 0),
                     up_frac_i=0.5, up_frac_f=0.5, up_theta=0., up_phi=0.):
    """
    *qx*, *qy* correspond to the detector pixels at which to calculate the
    scattering, relative to the beam along the negative z axis.
    *points* are three columns (x, y, z), one for each sample in the shape.
    *rho* (1e-6/Ang) is the scattering length density of each point.
    *rho_m* (1e-6/Ang) are the (mx, my, mz) components of the magnetic
    scattering length density for each point.
    *volume* should be 1/number_density.  That is, each of n particles in the
    total value represents volume/n contribution to the scattering.
    *view* rotates the points about the axes using Euler angles for pitch
    yaw and roll for a beam travelling along the negative z axis.
    *up_frac_i* is the portion of polarizer neutrons which are spin up.
    *up_frac_f* is the portion of analyzer neutrons which are spin up.
    *up_theta* is the inclination from the beam direction (z-axis).
    *up_phi* is the rotation in the detector plane.
    *dtype* is the numerical precision of the calculation. [not implemented]
    """
    # TODO: maybe slightly faster to rotate points and rho_m, and drop qc*z
    qx, qy = np.broadcast_arrays(qx, qy)
    qa, qb, qc = invert_view(qx, qy, view)
    rho, volume = np.broadcast_arrays(rho, volume)
    weights = spin_weights(up_frac_i, up_frac_f)

    # I(q) = |sum V(r) rho(r) e^(1j q.r)|^2 / sum V(r)
    shape = qx.shape
    Iq = np.zeros(qx.size, 'd')
    x, y, z = points.T
    qx, qy = (v.flatten() for v in (qx, qy))
    for k in range(qx.size):
        ephase = volume*np.exp(1j*(qa[k]*x + qb[k]*y + qc[k]*z))
        dd, du, ud, uu = magnetic_sld(qx[k], qy[k], up_theta, up_phi, rho, rho_m)
        for w, xs in zip(weights, (dd, du, ud, uu)):
            if w == 0.0:
                continue
            Iq[k] += w * abs(np.sum(xs*ephase))**2
    # The scale factor 1e-4 is due to the conversion from rho = 1e-6 squared
    # times the conversion of 1e-8 from inverse angstroms to inverse cm.
    return np.asarray(Iq).reshape(shape) * (1e-4 / np.sum(volume))

def _calc_Pr_nonuniform(r, rho, points, volume):
    # Make Pr a little be bigger than necessary so that only distances
    # min < d < max end up in Pr
    n_max = len(r)+1
    extended_Pr = np.zeros(n_max+1, 'd')
    # r refers to bin centers; find corresponding bin edges
    bins = bin_edges(r)
    t_next = timer() + 3
    for k, rho_k in enumerate(rho[:-1]):
        distance = np.linalg.norm(points[k] - points[k+1:], axis=1)
        weights = (rho_k * volume[k]) * (rho[k+1:] * volume[k+1:])
        #weights = (rho_k * volume[k]) * rho[k+1:]
        index = np.searchsorted(bins, distance)
        # Note: indices may be duplicated, so "Pr[index] += w" will not work!!
        extended_Pr += np.bincount(index, weights, n_max+1)
        t = timer()
        if t > t_next:
            t_next = t + 3
            print("processing %d of %d"%(k, len(rho)-1))
    Pr = extended_Pr[1:-1]
    return Pr

def _calc_Pr_uniform(r, rho, points, volume):
    # Make Pr a little be bigger than necessary so that only distances
    # min < d < max end up in Pr
    dr, n_max = r[0], len(r)
    extended_Pr = np.zeros(n_max+1, 'd')
    t0 = timer()
    t_next = t0 + 3
    for k, rho_k in enumerate(rho[:-1]):
        distance = np.linalg.norm(points[k] - points[k+1:], axis=1)
        #weights = (rho_k * volume[k]) * (rho[k+1:] * volume[k+1:])
        weights = rho_k * rho[k+1:] * (volume[k] + volume[k+1:])
        index = np.minimum(np.asarray(distance/dr, 'i'), n_max)
        # Note: indices may be duplicated, so "Pr[index] += w" will not work!!
        extended_Pr += np.bincount(index, weights, n_max+1)
        t = timer()
        if t > t_next:
            t_next = t + 3
            print("processing %d of %d"%(k, len(rho)-1))
    #print("time py:", timer() - t0)
    Pr = extended_Pr[:-1]
    #print("vol", np.sum(volume))
    return Pr*1e-4

    # Can get an additional 2x by going to C. Cuda/OpenCL will allow even
    # more speedup, though still bounded by the O(n^2) cost.
    """
void pdfcalc(int n, const double *pts, const double *rho,
  int nPr, double *Pr, double rstep)
{
  int i,j;
  for (i=0; i<n-2; i++) {
    for (j=i+1; j<=n-1; j++) {
      const double dxx=pts[3*i]-pts[3*j];
      const double dyy=pts[3*i+1]-pts[3*j+1];
      const double dzz=pts[3*i+2]-pts[3*j+2];
      const double d=sqrt(dxx*dxx+dyy*dyy+dzz*dzz);
      const int k=rint(d/rstep);
      if (k < nPr) Pr[k]+=rho[i]*rho[j];
    }
  }
}
"""

if USE_NUMBA:
    # Override simple numpy solution with numba if available
    #@njit("f8[:](f8[::1], f8[::1], f8[::1,:], f8[:])", parallel=True, fastmath=True)
    @njit(parallel=True, fastmath=True)
    def _calc_Pr_uniform(r, rho, points, volume):
        dr = r[0]
        n_max = len(r)
        Pr = np.zeros_like(r)
        for j in prange(len(rho) - 1):
            x, y, z = points[j, 0], points[j, 1], points[j, 2]
            rho_j, volume_j = rho[j], volume[j]
            for k in range(j+1, len(rho)):
                distance = sqrt((x - points[k, 0])**2
                                + (y - points[k, 1])**2
                                + (z - points[k, 2])**2)
                index = int(distance/dr)
                if index < n_max:
                    Pr[index] += rho_j*rho[k]*(volume_j + volume[k])
        return Pr


def calc_Pr(r, rho, points, volume):
    # P(r) with uniform steps in r is 3x faster; check if we are uniform
    # before continuing
    r, points = [np.asarray(v, 'd') for v in (r, points)]
    npoints = points.shape[0]
    rho = np.broadcast_to(np.asarray(rho, 'd'), npoints)
    volume = np.broadcast_to(np.asarray(volume, 'd'), npoints)
    if np.max(np.abs(np.diff(r) - r[0])) > r[0]*0.01:
        Pr = _calc_Pr_nonuniform(r, rho, points, volume)
    else:
        Pr = _calc_Pr_uniform(r, rho, points, volume)
    # Note: 1e-4 because (1e-6 rho)^2 = 1e-12 rho^2 time 1e-8 for 1/A to 1/cm
    return Pr * 1e-4


def r_bins(q, r_max=None, r_step=None, over_sampling=1):
    if r_max is None:
        r_max = 2 * pi / q[0]
    if r_step is None:
        r_step = 2 * pi / q[-1] / over_sampling
    return np.arange(r_step, r_max, r_step)


def j0(x):
    # use q/pi since np.sinc = sin(pi x)/(pi x)
    return np.sinc(x/np.pi)


def calc_Iq_from_Pr(q, r, Pr):
    Iq = np.array([simps(Pr * j0(qk*r), r) for qk in q])
    #Iq = np.array([np.trapz(Pr * j0(qk*r), r) for qk in q])
    #Iq /= Iq[0]
    return Iq


def _calc_Iq_avg(Iq, q, r, sld, volume):
    weight = sld * volume
    for i, qi in enumerate(q):
        Fq = np.sum(weight * np.sinc((qi/np.pi)*r))
        Iq[i] = Fq**2
if USE_NUMBA:
    #sig = njit('(f8[:], f8[:], f8[:], f8[:], f8[:])', parallel=True, fastmath=True)
    sig = njit(parallel=True, fastmath=True)
    _calc_Iq_avg = sig(_calc_Iq_avg)


def calc_Iq_avg(q, rho, points, volume=1.0):
    # Centralize the data
    center = 0.5*(np.min(points, axis=0, keepdims=True)
                  + np.max(points, axis=0, keepdims=True))
    points = points - center
    # Find distance from center
    r = np.linalg.norm(points, axis=1)
    # Call calculator
    Iq = np.empty_like(q)
    rho = np.broadcast_to(np.asarray(rho, 'd'), points.shape[:1])
    volume = np.broadcast_to(np.asarray(volume, 'd'), points.shape[:1])
    _calc_Iq_avg(Iq, q, r, rho, volume)
    return Iq * (1e-4/np.sum(volume))

# NOTE: copied from sasmodels/resolution.py
def bin_edges(x):
    """
    Determine bin edges from bin centers, assuming that edges are centered
    between the bins.
    Note: this uses the arithmetic mean, which may not be appropriate for
    log-scaled data.
    """
    if len(x) < 2 or (np.diff(x) < 0).any():
        raise ValueError("Expected bins to be an increasing set")
    edges = np.hstack([
        x[0]  - 0.5*(x[1]  - x[0]),  # first point minus half first interval
        0.5*(x[1:] + x[:-1]),        # mid points of all central intervals
        x[-1] + 0.5*(x[-1] - x[-2]), # last point plus half last interval
        ])
    return edges

# -------------- plotters ----------------
def plot_calc(r, Pr, q, Iq, theory=None, title=None, Iq_avg=None):
    import matplotlib.pyplot as plt
    plt.subplot(211)
    plt.plot(r, Pr, '-', label="Pr")
    plt.xlabel('r (A)')
    plt.ylabel('Pr (1/A^2)')
    if title is not None:
        plt.title(title)
    plt.grid(True)
    plt.subplot(212)
    plt.loglog(q, Iq, '-', label='from Pr')
    #plt.loglog(q, Iq/theory[1], '-', label='Pr/theory')
    if Iq_avg is not None:
        plt.loglog(q, Iq_avg, '-', label='from Iq_avg')
    plt.xlabel('q (1/A)')
    plt.ylabel('Iq')
    plt.grid(True)
    if theory is not None:
        #plt.loglog(theory[0], theory[1]/theory[1][0], '-', label='analytic')
        plt.loglog(theory[0], theory[1], '-', label='analytic')
        plt.legend()

def plot_calc_2d(qx, qy, Iqxy, theory=None, title=None):
    import matplotlib.pyplot as plt
    qx, qy = bin_edges(qx), bin_edges(qy)
    #qx, qy = np.meshgrid(qx, qy)
    if theory is not None:
        plt.subplot(131)
    #plt.pcolor(qx, qy, np.log10(Iqxy))
    extent = [qx[0], qx[-1], qy[0], qy[-1]]
    plt.imshow(np.log10(Iqxy), extent=extent, interpolation="nearest",
               origin='lower')
    plt.colorbar()
    plt.xlabel('qx (1/A)')
    plt.ylabel('qy (1/A)')
    plt.axis('equal')
    plt.axis(extent)
    #plt.grid(True)
    if title is not None:
        plt.title(title)
    if theory is not None:
        plt.subplot(132)
        # Skip bad values in theory
        index = np.isnan(theory)
        theory[index] = Iqxy[index]
        plt.imshow(np.log10(theory), extent=extent, interpolation="nearest",
                   origin='lower')
        plt.title("theory")
        plt.colorbar()
        plt.axis('equal')
        plt.axis(extent)
        plt.xlabel('qx (1/A)')

    if theory is not None:
        plt.subplot(133)
        rel = (theory-Iqxy)/theory
        plt.imshow(rel, extent=extent, interpolation="nearest", origin='lower')
        plt.colorbar()
        plt.axis('equal')
        plt.axis(extent)
        plt.xlabel('qx (1/A)')
        plt.title('max rel. err=%g' % np.max(abs(rel)))

def plot_points(rho, points):
    import mpl_toolkits.mplot3d
    import matplotlib.pyplot as plt

    ax = plt.axes(projection='3d')
    try:
        ax.axis('square')
    except Exception:
        pass
    n = len(points)
    #print("len points", n)
    index = np.random.choice(n, size=500) if n > 500 else slice(None, None)
    ax.scatter(points[index, 0], points[index, 1], points[index, 2], c=rho[index])
    # make square axes
    minmax = np.array([points.min(), points.max()])
    ax.scatter(minmax, minmax, minmax, c='w')
    #low, high = points.min(axis=0), points.max(axis=0)
    #ax.axis([low[0], high[0], low[1], high[1], low[2], high[2]])
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")
    ax.autoscale(True)

# ----------- Analytic models --------------
def sas_sinx_x(x):
    with np.errstate(all='ignore'):
        retvalue = sin(x)/x
    retvalue[x == 0.] = 1.
    return retvalue

def sas_2J1x_x(x):
    with np.errstate(all='ignore'):
        retvalue = 2*J1(x)/x
    retvalue[x == 0] = 1.
    return retvalue

def sas_3j1x_x(x):
    """return 3*j1(x)/x"""
    with np.errstate(all='ignore'):
        retvalue = 3*(sin(x) - x*cos(x))/x**3
    retvalue[x == 0.] = 1.
    return retvalue

def cylinder_Iq(q, radius, length):
    z, w = leggauss(76)
    cos_alpha = (z+1)/2
    sin_alpha = sqrt(1.0 - cos_alpha**2)
    Iq = np.empty_like(q)
    for k, qk in enumerate(q):
        qab, qc = qk*sin_alpha, qk*cos_alpha
        Fq = sas_2J1x_x(qab*radius) * sas_sinx_x(qc*length/2)
        Iq[k] = np.sum(w*Fq**2)
    Iq = Iq
    return Iq

def cylinder_Iqxy(qx, qy, radius, length, view=(0, 0, 0)):
    qa, qb, qc = invert_view(qx, qy, view)
    qab = sqrt(qa**2 + qb**2)
    Fq = sas_2J1x_x(qab*radius) * sas_sinx_x(qc*length/2)
    Iq = Fq**2
    return Iq.reshape(qx.shape)

def sphere_Iq(q, radius):
    Iq = sas_3j1x_x(q*radius)**2
    return Iq

def box_Iq(q, a, b, c):
    z, w = leggauss(76)
    outer_sum = np.zeros_like(q)
    for cos_alpha, outer_w in zip((z+1)/2, w):
        sin_alpha = sqrt(1.0-cos_alpha*cos_alpha)
        qc = q*cos_alpha
        siC = c*sas_sinx_x(c*qc/2)
        inner_sum = np.zeros_like(q)
        for beta, inner_w in zip((z + 1)*pi/4, w):
            qa, qb = q*sin_alpha*sin(beta), q*sin_alpha*cos(beta)
            siA = a*sas_sinx_x(a*qa/2)
            siB = b*sas_sinx_x(b*qb/2)
            Fq = siA*siB*siC
            inner_sum += inner_w * Fq**2
        outer_sum += outer_w * inner_sum
    Iq = outer_sum / 4  # = outer*um*zm*8.0/(4.0*M_PI)
    return Iq

def box_Iqxy(qx, qy, a, b, c, view=(0, 0, 0)):
    qa, qb, qc = invert_view(qx, qy, view)
    sia = sas_sinx_x(qa*a/2)
    sib = sas_sinx_x(qb*b/2)
    sic = sas_sinx_x(qc*c/2)
    Fq = sia*sib*sic
    Iq = Fq**2
    return Iq.reshape(qx.shape)

def csbox_Iq(q, a, b, c, da, db, dc, slda, sldb, sldc, sld_core):
    z, w = leggauss(76)

    sld_solvent = 0
    overlapping = False
    dr0 = sld_core - sld_solvent
    drA, drB, drC = slda-sld_solvent, sldb-sld_solvent, sldc-sld_solvent
    tA, tB, tC = a + 2*da, b + 2*db, c + 2*dc

    outer_sum = np.zeros_like(q)
    for cos_alpha, outer_w in zip((z+1)/2, w):
        sin_alpha = sqrt(1.0-cos_alpha*cos_alpha)
        qc = q*cos_alpha
        siC = c*sas_sinx_x(c*qc/2)
        siCt = tC*sas_sinx_x(tC*qc/2)
        inner_sum = np.zeros_like(q)
        for beta, inner_w in zip((z + 1)*pi/4, w):
            qa, qb = q*sin_alpha*sin(beta), q*sin_alpha*cos(beta)
            siA = a*sas_sinx_x(a*qa/2)
            siB = b*sas_sinx_x(b*qb/2)
            siAt = tA*sas_sinx_x(tA*qa/2)
            siBt = tB*sas_sinx_x(tB*qb/2)
            if overlapping:
                Fq = (dr0*siA*siB*siC
                      + drA*(siAt-siA)*siB*siC
                      + drB*siAt*(siBt-siB)*siC
                      + drC*siAt*siBt*(siCt-siC))
            else:
                Fq = (dr0*siA*siB*siC
                      + drA*(siAt-siA)*siB*siC
                      + drB*siA*(siBt-siB)*siC
                      + drC*siA*siB*(siCt-siC))
            inner_sum += inner_w * Fq**2
        outer_sum += outer_w * inner_sum
    Iq = outer_sum / 4  # = outer*um*zm*8.0/(4.0*M_PI)
    return Iq/Iq[0]

def csbox_Iqxy(qx, qy, a, b, c, da, db, dc, slda, sldb, sldc, sld_core, view=(0,0,0)):
    qa, qb, qc = invert_view(qx, qy, view)

    sld_solvent = 0
    overlapping = False
    dr0 = sld_core - sld_solvent
    drA, drB, drC = slda-sld_solvent, sldb-sld_solvent, sldc-sld_solvent
    tA, tB, tC = a + 2*da, b + 2*db, c + 2*dc
    siA = a*sas_sinx_x(a*qa/2)
    siB = b*sas_sinx_x(b*qb/2)
    siC = c*sas_sinx_x(c*qc/2)
    siAt = tA*sas_sinx_x(tA*qa/2)
    siBt = tB*sas_sinx_x(tB*qb/2)
    siCt = tC*sas_sinx_x(tC*qc/2)
    Fq = (dr0*siA*siB*siC
          + drA*(siAt-siA)*siB*siC
          + drB*siA*(siBt-siB)*siC
          + drC*siA*siB*(siCt-siC))
    Iq = Fq**2
    return Iq.reshape(qx.shape)

def _sasmodels_Iq(kernel, q, pars):
    from sasmodels.data import empty_data1D
    from sasmodels.direct_model import DirectModel
    data = empty_data1D(q)
    calculator = DirectModel(data, kernel)
    Iq = calculator(**pars)
    return Iq

def _sasmodels_Iqxy(kernel, qx, qy, pars, view):
    from sasmodels.data import Data2D
    from sasmodels.direct_model import DirectModel
    Iq = np.full_like(qx, 100)
    data = Data2D(x=qx, y=qy, z=Iq, dx=None, dy=None, dz=np.sqrt(Iq))
    data.x_bins = qx[0, :]
    data.y_bins = qy[:, 0]
    data.filename = "fake data"

    calculator = DirectModel(data, kernel)
    pars_plus_view = pars.copy()
    pars_plus_view.update(theta=view[0], phi=view[1], psi=view[2])
    Iqxy = calculator(**pars_plus_view)
    # calculator avoids masked values; instead set masked values to NaN
    result = np.empty_like(qx)
    result[calculator.index] = Iqxy
    result[~calculator.index] = np.nan
    return result

def wrap_sasmodel(name, **pars):
    from sasmodels.core import load_model
    kernel = load_model(name)
    fn = lambda q: _sasmodels_Iq(kernel, q, pars)
    fn_xy = lambda qx, qy, view: _sasmodels_Iqxy(kernel, qx, qy, pars, view)
    return fn, fn_xy


# --------- Test cases -----------

def build_box(a=10, b=20, c=30, rho=2.):
    shape = Box(a, b, c, rho)
    fn = lambda q: box_Iq(q, a, b, c)*rho**2
    fn_xy = lambda qx, qy, view: box_Iqxy(qx, qy, a, b, c, view=view)*rho**2
    return shape, fn, fn_xy

def build_superball(a=10, p=3, rho=2.):
    shape = Superball(a, p, rho)
    fn, fn_xy = wrap_sasmodel(
        'superball',
        scale=1,
        background=0,
        length_a=a,
        exponent_p=p,
        sld=rho,
        sld_solvent=0,
    )
    return shape, fn, fn_xy

def build_csbox(a=10, b=20, c=30, da=1, db=2, dc=3, slda=1, sldb=2, sldc=3, sld_core=4):
    shape = csbox(a, b, c, da, db, dc, slda, sldb, sldc, sld_core)
    fn = lambda q: csbox_Iq(q, a, b, c, da, db, dc, slda, sldb, sldc, sld_core)
    fn_xy = lambda qx, qy, view: csbox_Iqxy(qx, qy, a, b, c, da, db, dc,
                                            slda, sldb, sldc, sld_core, view=view)
    return shape, fn, fn_xy

def build_sphere(radius=125, rho=2,
                 rho_m=0, theta_m=0, phi_m=0, up_i=0, up_f=0, up_theta=0, up_phi=0):
    magnetism = pol2rec(rho_m, theta_m, phi_m) if rho_m != 0.0 else None
    shape = TriaxialEllipsoid(radius, radius, radius, rho, magnetism=magnetism)
    shape.spin = (up_i, up_f, up_theta, up_phi)
    fn, fn_xy = wrap_sasmodel(
        'sphere',
        scale=1,
        background=0,
        radius=radius,
        sld=rho,
        sld_solvent=0,
        sld_M0=rho_m,
        sld_mtheta=theta_m,
        sld_mphi=phi_m,
        up_frac_i=up_i,
        up_frac_f=up_f,
        up_theta=up_theta,
        up_phi=up_phi,
    )
    return shape, fn, fn_xy

def build_ellip(rab=125, rc=50, rho=2,
                rho_m=0, theta_m=0, phi_m=0, up_i=0, up_f=0, up_theta=0, up_phi=0):
    magnetism = pol2rec(rho_m, theta_m, phi_m) if rho_m != 0.0 else None
    shape = TriaxialEllipsoid(rab, rab, rc, rho, magnetism=magnetism)
    # TODO: polarization spec doesn't belong in shape
    # Put spin state info into the shape since we have it available.
    shape.spin = (up_i, up_f, up_theta, up_phi)
    fn, fn_xy = wrap_sasmodel(
        'ellipsoid',
        scale=1,
        background=0,
        radius_equatorial=rab,
        radius_polar=rc,
        sld=rho,
        sld_solvent=0,
        sld_M0=rho_m,
        sld_mtheta=theta_m,
        sld_mphi=phi_m,
        up_frac_i=up_i,
        up_frac_f=up_f,
        up_theta=up_theta,
        up_phi=up_phi,
    )
    return shape, fn, fn_xy

def build_triell(ra=125, rb=200, rc=50, rho=2,
                 rho_m=0, theta_m=0, phi_m=0, up_i=0, up_f=0, up_theta=0, up_phi=0):
    magnetism = pol2rec(rho_m, theta_m, phi_m) if rho_m != 0.0 else None
    shape = TriaxialEllipsoid(ra, rb, rc, rho, magnetism=magnetism)
    shape.spin = (up_i, up_f, up_theta, up_phi)
    fn, fn_xy = wrap_sasmodel(
        'triaxial_ellipsoid',
        scale=1,
        background=0,
        radius_equat_minor=ra,
        radius_equat_major=rb,
        radius_polar=rc,
        sld=rho,
        sld_solvent=0,
        sld_M0=rho_m,
        sld_mtheta=theta_m,
        sld_mphi=phi_m,
        up_frac_i=up_i,
        up_frac_f=up_f,
        up_theta=up_theta,
        up_phi=up_phi,
    )
    return shape, fn, fn_xy

def build_cylinder(radius=25, length=125, rho=2.):
    shape = EllipticalCylinder(radius, radius, length, rho)
    fn = lambda q: cylinder_Iq(q, radius, length)*rho**2
    fn_xy = lambda qx, qy, view: cylinder_Iqxy(qx, qy, radius, length, view=view)*rho**2
    return shape, fn, fn_xy

def build_truncated_sphere(radius=25, h=0.5, rho=2.):
    shape = TruncatedSphere(radius, h=radius*h, value=rho)
    return shape, None, None

def build_ellcyl(ra=25, rb=50, length=125, rho=2.):
    shape = EllipticalCylinder(ra, rb, length, rho)
    fn, fn_xy = wrap_sasmodel(
        'elliptical_cylinder',
        scale=1,
        background=0,
        radius_minor=ra,
        axis_ratio=rb/ra,
        length=length,
        sld=rho,
        sld_solvent=0,
    )
    return shape, fn, fn_xy

def build_barbell(r=20, rbell=50, length=20, rho=2.):
    shape = barbell(r, rbell, length, rho)
    fn, fn_xy = wrap_sasmodel(
        'barbell',
        scale=1,
        background=0,
        radius=r,
        radius_bell=rbell,
        length=length,
        sld=rho,
        sld_solvent=0,
    )
    return shape, fn, fn_xy

def build_capcyl(r=20, rcap=50, length=20, rho=2.):
    shape = capped_cylinder(r, rcap, length, rho)
    fn, fn_xy = wrap_sasmodel(
        'capped_cylinder',
        scale=1,
        background=0,
        radius=r,
        radius_cap=rcap,
        sld=rho,
        sld_solvent=0,
    )
    return shape, fn, fn_xy

def build_cscyl(ra=30, rb=90, length=30, thick_rim=8, thick_face=14,
                sld_core=4, sld_rim=1, sld_face=7):
    shape = EllipticalBicelle(
        ra=ra, rb=rb, length=length,
        thick_rim=thick_rim, thick_face=thick_face,
        value_core=sld_core, value_rim=sld_rim, value_face=sld_face,
        )
    fn, fn_xy = wrap_sasmodel(
        'core_shell_bicelle_elliptical',
        scale=1,
        background=0,
        radius=ra,
        x_core=rb/ra,
        length=length,
        thick_rim=thick_rim,
        thick_face=thick_face,
        sld_core=sld_core,
        sld_face=sld_face,
        sld_rim=sld_rim,
        sld_solvent=0,
    )
    return shape, fn, fn_xy

def build_cubic_lattice(shape, nx=1, ny=1, nz=1, dx=2, dy=2, dz=2,
                        shuffle=0, rotate=0):
    a, b, c = shape.dims
    shapes = [copy(shape)
              .shift((ix+(randn() if shuffle < 0.3 else rand())*shuffle)*dx*a,
                     (iy+(randn() if shuffle < 0.3 else rand())*shuffle)*dy*b,
                     (iz+(randn() if shuffle < 0.3 else rand())*shuffle)*dz*c)
              .rotate(*((randn(3) if rotate < 30 else rand(3))*rotate))
              for ix in range(nx)
              for iy in range(ny)
              for iz in range(nz)]
    lattice = Composite(shapes)
    return lattice

SHAPE_FUNCTIONS = OrderedDict([
    ("cyl", build_cylinder),
    ("ellcyl", build_ellcyl),
    ("ellip", build_ellip),
    ("triell", build_triell),
    ("barbell", build_barbell),
    ("capcyl", build_capcyl),
    ("sphere", build_sphere),
    ("tsphere", build_truncated_sphere),
    ("box", build_box),
    ("csbox", build_csbox),
    ("cscyl", build_cscyl),
    ("superball", build_superball),
])
SHAPES = list(SHAPE_FUNCTIONS.keys())

def check_shape(title, shape, fn=None, show_points=False,
                mesh=100, qmax=1.0, r_step=0.01, samples=5000):
    rho_solvent = 0
    qmin = qmax/100.
    q = np.logspace(np.log10(qmin), np.log10(qmax), mesh)
    r = shape.r_bins(q, r_step=r_step)
    sampling_density = samples / shape.volume
    rho, points = shape.sample(sampling_density)
    volume = shape.volume / len(points)
    t0 = timer()
    Pr = calc_Pr(r, rho-rho_solvent, points, volume)
    print("calc Pr time", timer() - t0)
    Iq = calc_Iq_from_Pr(q, r, Pr)
    t0 = timer()
    Iq_avg = calc_Iq_avg(q, rho-rho_solvent, points, volume)
    print("calc Iq_avg time", timer() - t0)
    theory = (q, fn(q)) if fn is not None else None

    import pylab
    if show_points:
        plot_points(rho, points)
        pylab.figure()
    plot_calc(r, Pr, q, Iq, theory=theory, title=title, Iq_avg=Iq_avg)
    pylab.gcf().canvas.manager.set_window_title(title)
    pylab.show()

def check_shape_2d(title, shape, fn=None, view=(0, 0, 0), show_points=False,
                   mesh=100, qmax=1.0, samples=5000):
    rho_solvent = 0
    #qx = np.linspace(0.0, qmax, mesh)
    #qy = np.linspace(0.0, qmax, mesh)
    qx = np.linspace(-qmax, qmax, mesh)
    qy = np.linspace(-qmax, qmax, mesh)
    Qx, Qy = np.meshgrid(qx, qy)
    t0 = timer()
    theory = fn(Qx, Qy, view) if fn is not None else None
    print("calc theory time", timer() - t0)

    t0 = timer()
    sampling_density = samples / shape.volume
    if False: # point orientation test: rotate shape rather than view
        shape.rotate(*view)
        view = (0, 0, 0)
    rho, points = shape.sample(sampling_density)
    # The volume of each sample is approximately 1/sampling_density, except
    # that the number of points actually sampled may be slightly more or
    # less due to nature of how points are sampled from the shape.
    volume = shape.volume / len(rho)
    print("point generation time", timer() - t0)
    #print("bounding box", points.min(axis=0), points.max(axis=0))
    # Call calculator with a trivial problem to hide compile times.
    calc_Iqxy(Qx[:1,:1], Qy[:1,:1], rho[:1], points[:1])
    t0 = timer()
    Iqxy = calc_Iqxy(Qx, Qy, rho, points, volume=volume, view=view)
    print("calc Iqxy time", timer() - t0)

    # Add floor to limit colorbar range.
    Iqxy += 0.001 * Iqxy.max()
    if theory is not None:
        theory += 0.001 * theory[~np.isnan(theory)].max()

    import pylab
    if show_points:
        plot_points(rho, apply_view(points, view))
        pylab.figure()
    plot_calc_2d(qx, qy, Iqxy, theory=theory, title=title)
    pylab.gcf().canvas.manager.set_window_title(title)

    ## Histogram of point density in the z direction.
    #pylab.figure()
    ##bins = 100  # fixed number of bins
    #limit=25; bins = np.arange(-limit, limit+0.001, 1) # particular limits
    #pylab.hist(points[:,2], bins=bins)

    pylab.show()

def check_shape_mag(title, shape, fn=None, view=(0, 0, 0), show_points=False,
                    mesh=100, qmax=1.0, samples=5000,
                    up_frac_i=0, up_frac_f=0, up_theta=0, up_phi=0):
    rho_solvent = 0
    #qx = np.linspace(0.0, qmax, mesh)
    #qy = np.linspace(0.0, qmax, mesh)
    qx = np.linspace(-qmax, qmax, mesh)
    qy = np.linspace(-qmax, qmax, mesh)
    Qx, Qy = np.meshgrid(qx, qy)
    sampling_density = samples / shape.volume
    t0 = timer()
    rho, rho_m, points = shape.sample_magnetic(sampling_density)
    # The volume of each sample is approximately 1/sampling_density, except
    # that the number of points actually sampled may be slightly more or
    # less due to nature of how points are sampled from the shape.
    volume = shape.volume / len(rho)
    print("point generation time", timer() - t0)
    #print("bounding box", points.min(axis=0), points.max(axis=0))
    t0 = timer()
    Iqxy = calc_Iq_magnetic(Qx, Qy, rho, rho_m, points, volume=volume, view=view,
                            up_frac_i=up_frac_i, up_frac_f=up_frac_f,
                            up_theta=up_theta, up_phi=up_phi)
    print("calc I_mag time", timer() - t0)
    t0 = timer()
    theory = fn(Qx, Qy, view) if fn is not None else None
    print("calc theory time", timer() - t0)

    # Add floor to limit colorbar range.
    Iqxy += 0.001 * Iqxy.max()
    if theory is not None:
        theory += 0.001 * theory.max()

    import pylab
    if show_points:
        plot_points(rho, points)
        pylab.figure()
    plot_calc_2d(qx, qy, Iqxy, theory=theory, title=title)
    pylab.gcf().canvas.manager.set_window_title(title)

    ## Histogram of point density in the z direction.
    #pylab.figure()
    ##bins = 100  # fixed number of bins
    #limit=25; bins = np.arange(-limit, limit+0.001, 1) # particular limits
    #pylab.hist(points[:,2], bins=bins)

    pylab.show()


def check_pars(function, pars, name):
    spec = getfullargspec(function)
    incorrect = [p for p in pars if p not in spec.args]
    if any(incorrect):
        print("realspace.py: error: invalid parameters to %s: [%s].  Available paramters are:"
              % (name, ", ".join(incorrect)))
        print("   ", "\n    ".join("%s: %g"%(k,v) for k,v in zip(spec.args, spec.defaults)))
        return False
    return True

def main():
    parser = argparse.ArgumentParser(
        description="Compute scattering from realspace sampling",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
    parser.add_argument('-d', '--dim', type=int, default=1,
                        help='dimension 1 or 2')
    parser.add_argument('-m', '--mesh', type=int, default=100,
                        help='number of mesh points in the computed scattering')
    parser.add_argument('-s', '--samples', type=int, default=5000,
                        help="number of sample points in the Monte Carlo estimate")
    parser.add_argument('-q', '--qmax', type=float, default=0.5,
                        help='max q')
    parser.add_argument('-v', '--view', type=str, default='0,0,0',
                        help='theta,phi,psi angles')
    parser.add_argument('-n', '--lattice', type=str, default='1,1,1',
                        help='lattice size')
    parser.add_argument('-z', '--spacing', type=str, default='2,2,2',
                        help='lattice spacing')
    parser.add_argument('-r', '--rotate', type=float, default=0.,
                        help="rotation relative to lattice, gaussian < 30 degrees, uniform otherwise")
    parser.add_argument('-w', '--shuffle', type=float, default=0.,
                        help="position relative to lattice, gaussian < 0.3, uniform otherwise")
    parser.add_argument('-p', '--plot', action='store_true',
                        help='plot points')
    parser.add_argument('shape', choices=SHAPES, nargs='?', default=SHAPES[0],
                        help='oriented shape')
    parser.add_argument('pars', type=str, nargs='*', help='shape parameters')
    opts = parser.parse_args()
    pars = {key: float(value) for p in opts.pars for key, value in [p.split('=')]}
    nx, ny, nz = [int(v) for v in opts.lattice.split(',')]
    dx, dy, dz = [float(v) for v in opts.spacing.split(',')]
    shuffle, rotate = opts.shuffle, opts.rotate
    shape_generator = SHAPE_FUNCTIONS[opts.shape]
    if not check_pars(shape_generator, pars, name=opts.shape):
        return
    shape, fn, fn_xy = shape_generator(**pars)
    if nx > 1 or ny > 1 or nz > 1:
        shape = build_cubic_lattice(shape, nx, ny, nz, dx, dy, dz, shuffle, rotate)
    title = "%s(%s)" % (opts.shape, " ".join(opts.pars))
    if shape.is_magnetic:
        view = tuple(float(v) for v in opts.view.split(','))
        up_frac_i, up_frac_f, up_theta, up_phi = shape.spin
        check_shape_mag(title, shape, fn_xy, view=view, show_points=opts.plot,
                       mesh=opts.mesh, qmax=opts.qmax, samples=opts.samples,
                       up_frac_i=up_frac_i, up_frac_f=up_frac_f, up_theta=up_theta, up_phi=up_phi,
                       )
    elif opts.dim == 1:
        check_shape(title, shape, fn, show_points=opts.plot,
                    mesh=opts.mesh, qmax=opts.qmax, samples=opts.samples)
    else:
        view = tuple(float(v) for v in opts.view.split(','))
        check_shape_2d(title, shape, fn_xy, view=view, show_points=opts.plot,
                       mesh=opts.mesh, qmax=opts.qmax, samples=opts.samples)


if __name__ == "__main__":
    # Make sure sasmodels in on the path
    try:
        import sasmodels
    except ImportError:
        import sys
        from os.path import realpath, dirname, join as joinpath
        sys.path.insert(0, dirname(dirname(realpath(__file__))))
    main()



r"""
Definition
----------

This model provides the form factor for an elliptical cylinder with a
core-shell scattering length density profile [#Onsager1949]_.
Thus this is a variation
of the core-shell bicelle model, but with an elliptical cylinder for the core.
In this version the "rim" or "belt" does NOT extend the full length of
the particle, but has the same length as the core. Outer shells on the
rims and flat ends may be of different thicknesses and scattering length
densities. The form factor is normalized by the total particle volume.
This version includes an approximate "interfacial roughness".


.. figure:: img/core_shell_bicelle_belt_geometry.png

    Schematic cross-section of bicelle with belt. Note however that the model
    here calculates for rectangular, not curved, rims as shown below.

.. figure:: img/core_shell_bicelle_belt_parameters.png

   Cross section of model used here. Users will have
   to decide how to distribute "heads" and "tails" between the rim, face
   and core regions in order to estimate appropriate starting parameters.

Given the scattering length densities (sld) $\rho_c$, the core sld, $\rho_f$,
the face sld, $\rho_r$, the rim sld and $\rho_s$ the solvent sld, the
scattering length density variation along the bicelle axis is:

.. math::

    \rho(r) =
      \begin{cases}
      &\rho_c \text{ for } 0 \lt r \lt R;   -L/2 \lt z\lt L/2 \\[1.5ex]
      &\rho_f \text{ for } 0 \lt r \lt R;   -(L/2 +t_\text{face}) \lt z\lt -L/2;
      L/2 \lt z\lt (L/2+t_\text{face}) \\[1.5ex]
      &\rho_r\text{ for } R \lt r \lt R+t_\text{rim}; -L/2 \lt z\lt L/2
      \end{cases}

The form factor for the bicelle is calculated in cylindrical coordinates, where
$\alpha$ is the angle between the $Q$ vector and the cylinder axis, and $\psi$
is the angle for the ellipsoidal cross section core, to give:

.. math::

    I(Q,\alpha,\psi) = \frac{\text{scale}}{V_t}
        \cdot F(Q,\alpha, \psi)^2 \cdot \sin(\alpha)
        \cdot\exp\left\{ -\frac{1}{2}Q^2\sigma^2 \right\} + \text{background}

where a numerical integration of $F(Q,\alpha, \psi)^2\sin(\alpha)$ is
carried out over $\alpha$ and $\psi$ for:

.. math::

    F(Q,\alpha,\psi) = &\bigg[
      (\rho_c -\rho_r - \rho_f + \rho_s) V_c
      \frac{2J_1(QR'\sin \alpha)}{QR'\sin\alpha}
      \frac{\sin(QL\cos\alpha/2)}{Q(L/2)\cos\alpha} \\
    &+(\rho_f - \rho_s) V_{c+f}
      \frac{2J_1(QR'\sin\alpha)}{QR'\sin\alpha}
      \frac{\sin(Q(L/2+t_f)\cos\alpha)}{Q(L/2+t_f)\cos\alpha} \\
    &+(\rho_r - \rho_s) V_{c+r}
      \frac{2J_1(Q(R'+t_r)\sin\alpha)}{Q(R'+t_r)\sin\alpha}
      \frac{\sin(Q(L/2)\cos\alpha)}{Q(L/2)\cos\alpha}
    \bigg]

where

.. math::

    R' = \frac{R}{\sqrt{2}}
        \sqrt{(1+X_\text{core}^{2}) + (1-X_\text{core}^{2})\cos(\psi)}


and $V_t = \pi (R+t_r)(X_\text{core} R+t_r) L + 2 \pi X_\text{core} R^2 t_f$ is
the total volume of the bicelle, $V_c = \pi X_\text{core} R^2 L$ the volume of
the core, $V_{c+f} = \pi X_\text{core} R^2 (L+2 t_f)$ the volume of the core
plus the volume of the faces, $V_{c+r} = \pi (R+t_r)(X_\text{core} R+t_r) L$
the volume of the core plus the rim, $R$ is the radius of the core,
$X_\text{core}$ is the axial ratio of the core, $L$ the length of the core,
$t_f$ the thickness of the face, $t_r$ the thickness of the rim and $J_1$ the
usual first order Bessel function. The core has radii $R$ and $X_\text{core} R$
so is circular, as for the core_shell_bicelle model, for $X_\text{core}=1$.
Note that you may need to limit the range of $X_\text{core}$, especially if
using the Monte-Carlo algorithm, as setting radius to $R/X_\text{core}$ and
axial ratio to $1/X_\text{core}$ gives an equivalent solution!

An approximation for the effects of "Gaussian interfacial roughness" $\sigma$
is included, by multiplying $I(Q)$ by
$\exp\left \{ -\frac{1}{2}Q^2\sigma^2 \right \}$. This applies, in some way, to
all interfaces in the model not just the external ones. (Note that for a one
dimensional system convolution of the scattering length density profile with
a Gaussian of standard deviation $\sigma$ does exactly this multiplication.)
Leave $\sigma$ set to zero for the usual sharp interfaces.

The output of the 1D scattering intensity function for randomly oriented
bicelles is then given by integrating over all possible $\alpha$ and $\psi$.

For oriented bicelles the *theta*, *phi* and *psi* orientation parameters
will appear when fitting 2D data, for further details of the calculation
and angular dispersions  see :ref:`orientation` .

.. figure:: img/elliptical_cylinder_angle_definition.png

    Definition of the angles for the oriented core_shell_bicelle_elliptical
    particles.



References
----------

.. [#Onsager1949] L. Onsager, *Ann. New York Acad. Sci.*, 51 (1949) 627-659

Authorship and Verification
----------------------------

* **Author:** Richard Heenan **Date:** October 5, 2017
* **Last Modified by:**  Richard Heenan new 2d orientation **Date:** October 5, 2017
* **Last Reviewed by:**  Richard Heenan 2d calc seems agree with 1d **Date:** Nov 2, 2017
"""

from numpy import inf, sin, cos, pi

name = "core_shell_bicelle_elliptical_belt_rough"
title = "Elliptical cylinder with a core-shell scattering length density profile.."
description = """
    core_shell_bicelle_elliptical_belt_rough
    Elliptical cylinder core, optional shell on the two flat faces, and "belt" shell of
    uniform thickness on its rim (in this case NOT extending around the end faces).
    with approximate interfacial roughness.
    Please see full documentation for equations and further details.
    Involves a double numerical integral around the ellipsoid diameter
    and the angle of the cylinder axis to Q.
    Compare also the core_shell_bicelle and elliptical_cylinder models.
      """
category = "shape:cylinder"

# pylint: disable=bad-whitespace, line-too-long
#             ["name", "units", default, [lower, upper], "type", "description"],
parameters = [
    ["radius",         "Ang",       30, [0, inf],    "volume",      "Cylinder core radius r_minor"],
    ["x_core",        "None",       3,  [0, inf],    "volume",      "Axial ratio of core, X = r_major/r_minor"],
    ["thick_rim",  "Ang",            8, [0, inf],    "volume",      "Rim or belt shell thickness"],
    ["thick_face", "Ang",           14, [0, inf],    "volume",      "Cylinder face thickness"],
    ["length",         "Ang",       50, [0, inf],    "volume",      "Cylinder length"],
    ["sld_core",       "1e-6/Ang^2", 4, [-inf, inf], "sld",         "Cylinder core scattering length density"],
    ["sld_face",       "1e-6/Ang^2", 7, [-inf, inf], "sld",         "Cylinder face scattering length density"],
    ["sld_rim",        "1e-6/Ang^2", 1, [-inf, inf], "sld",         "Cylinder rim scattering length density"],
    ["sld_solvent",    "1e-6/Ang^2", 6, [-inf, inf], "sld",         "Solvent scattering length density"],
    ["sigma",       "Ang",        0,    [0, inf],    "",            "Interfacial roughness"],
    ["theta",       "degrees",    90.0, [-360, 360], "orientation", "Cylinder axis to beam angle"],
    ["phi",         "degrees",    0,    [-360, 360], "orientation", "Rotation about beam"],
    ["psi",         "degrees",    0,    [-360, 360], "orientation", "Rotation about cylinder axis"],
    ]

# pylint: enable=bad-whitespace, line-too-long

source = ["lib/sas_Si.c", "lib/polevl.c", "lib/sas_J1.c", "lib/gauss76.c",
          "core_shell_bicelle_elliptical_belt_rough.c"]
have_Fq = True
radius_effective_modes = [
    "equivalent cylinder excluded volume", "equivalent volume sphere",
    "outer rim average radius", "outer rim min radius",
    "outer max radius", "half outer thickness", "half diagonal",
    ]

# TODO: No random() for core-shell bicelle elliptical belt rough

q = 0.1
# april 6 2017, rkh added a 2d unit test, NOT READY YET pull #890 branch assume correct!
qx = q*cos(pi/6.0)
qy = q*sin(pi/6.0)

tests = [
    #[{'radius': 30.0, 'x_core': 3.0, 'thick_rim':8.0, 'thick_face':14.0, 'length':50.0}, 'ER', 1],
    #[{'radius': 30.0, 'x_core': 3.0, 'thick_rim':8.0, 'thick_face':14.0, 'length':50.0}, 'VR', 1],

    [{'radius': 30.0, 'x_core': 3.0, 'thick_rim': 8.0, 'thick_face': 14.0,
      'length': 50.0, 'sld_core': 4.0, 'sld_face': 7.0, 'sld_rim': 1.0,
      'sld_solvent': 6.0, 'background': 0.0},
     0.015, 189.328],
    #[{'theta':80., 'phi':10.}, (qx, qy), 7.88866563001 ],
]

del qx, qy  # not necessary to delete, but cleaner



"""
Model definition files
----------------------

The models below are grouped by type. The documentation build system collects
and organizes the models automatically, so this list may be out of date.

Models with pure form factor (all of which define *F(Q)*):

    :mod:`barbell`
    :mod:`capped_cylinder`
    :mod:`core_multi_shell`
    :mod:`core_shell_bicelle`
    :mod:`core_shell_bicelle_elliptical`
    :mod:`core_shell_bicelle_elliptical_belt_rough`
    :mod:`core_shell_cylinder`
    :mod:`core_shell_ellipsoid`
    :mod:`core_shell_parallelepipied`
    :mod:`core_shell_sphere`
    :mod:`cylinder` [limiting conditions (long rods, thin disks) not implemented]
    :mod:`ellipsoid`
    :mod:`elliptical_cylinder`
    :mod:`fuzzy_sphere`
    :mod:`hollow_cylinder`
    :mod:`hollow_rectangular_prism`
    :mod:`hollow_rectangular_prism_thin_walls`
    :mod:`multilayer_vesicle`
    :mod:`onion`
    :mod:`parallelepiped`
    :mod:`rectangular_prism`
    :mod:`sphere`
    :mod:`spherical_sld`
    :mod:`triaxial_ellipsoid`
    :mod:`vesicle`

Models with local structure factor:

    :mod:`flexible_cylinder`
    :mod:`flexible_cylinder_elliptical`
    :mod:`linear_pearls`
    :mod:`mono_gauss_coil`
    :mod:`pearl_necklace`
    :mod:`poly_gauss_coil`
    :mod:`polymer_micelle`
    :mod:`pringle`
    :mod:`raspberry`
    :mod:`stacked_disks`
    :mod:`star_polymer`

Models with long range structure factor:

    :mod:`binary_hard_sphere`
    :mod:`bcc_paracrystal`
    :mod:`fcc_paracrystal`
    :mod:`fractal`
    :mod:`fractal_core_shell`
    :mod:`lamellar`
    :mod:`lamellar_hg`
    :mod:`lamellar_hg_stack_caille`
    :mod:`lamellar_stack_caille`
    :mod:`lamellar_stack_paracrystal`
    :mod:`mass_fractal`
    :mod:`mass_surface_fractal`
    :mod:`rpa`
    :mod:`sc_paracrystal`
    :mod:`surface_fractal`

Models which are pure structure factors::

    :mod:`hardsphere`
    :mod:`hayter_msa`
    :mod:`squarewell`
    :mod:`stickyhardsphere`

Other models:

    :mod:`adsorbed_layer`
    :mod:`be_polyelectrolyte`
    :mod:`broad_peak`
    :mod:`correlation_length`
    :mod:`dab`
    :mod:`gauss_lorentz_gel`
    :mod:`gaussian_peak`
    :mod:`gel_fit`
    :mod:`guinier_porod`
    :mod:`guinier`
    :mod:`line`
    :mod:`lorentz`
    :mod:`peak_lorentz`
    :mod:`polymer_excl_volume`
    :mod:`porod`
    :mod:`power_law`
    :mod:`spinodal`
    :mod:`teubner_strey`
    :mod:`two_lorentzian`
    :mod:`unified_power_Rg`

"""



r"""
Definition
----------

The output of the 2D scattering intensity function for oriented core-shell
cylinders is given by Kline [#Kline2006]_. The form factor is normalized
by the particle volume. Note that in this model the shell envelops the entire
core so that besides a "sleeve" around the core, the shell also provides two
flat end caps of thickness = shell thickness. In other words the length of the
total cylinder is the length of the core cylinder plus twice the thickness of
the shell. If no end caps are desired one should use the
:ref:`core-shell-bicelle` and set the thickness of the end caps (in this case
the "thick_face") to zero.

.. math::

    I(q,\alpha) = \frac{\text{scale}}{V_s} F^2(q,\alpha).sin(\alpha) + \text{background}

where

.. math::

    F(q,\alpha) = &\ (\rho_c - \rho_s) V_c
           \frac{\sin \left( q \tfrac12 L\cos\alpha \right)}
                {q \tfrac12 L\cos\alpha}
           \frac{2 J_1 \left( qR\sin\alpha \right)}
                {qR\sin\alpha} \\
         &\ + (\rho_s - \rho_\text{solv}) V_s
           \frac{\sin \left( q \left(\tfrac12 L+T\right) \cos\alpha \right)}
                {q \left(\tfrac12 L +T \right) \cos\alpha}
           \frac{ 2 J_1 \left( q(R+T)\sin\alpha \right)}
                {q(R+T)\sin\alpha}

and

.. math::

    V_s = \pi (R + T)^2 (L + 2T)

and $\alpha$ is the angle between the axis of the cylinder and $\vec q$,
$V_s$ is the total volume (i.e. including both the core and the outer shell),
$V_c$ is the volume of the core, $L$ is the length of the core,
$R$ is the radius of the core, $T$ is the thickness of the shell, $\rho_c$
is the scattering length density of the core, $\rho_s$ is the scattering
length density of the shell, $\rho_\text{solv}$ is the scattering length
density of the solvent, and *background* is the background level.  The outer
radius of the shell is given by $R+T$ and the total length of the outer
shell is given by $L+2T$. $J_1$ is the first order Bessel function.

.. _core-shell-cylinder-geometry:

.. figure:: img/core_shell_cylinder_geometry.jpg

    Core shell cylinder schematic.

To provide easy access to the orientation of the core-shell cylinder, we
define the axis of the cylinder using two angles $\theta$ and $\phi$.
(see :ref:`cylinder model <cylinder-angle-definition>`)

NB: The 2nd virial coefficient of the cylinder is calculated based on
the radius and 2 length values, and used as the effective radius for
$S(q)$ when $P(q) \cdot S(q)$ is applied.

The $\theta$ and $\phi$ parameters are not used for the 1D output.

Reference
---------

See also Livsey [#Livsey1987]_ and Onsager [#Onsager1949]_.

.. [#Livsey1987] I Livsey, *J. Chem. Soc., Faraday Trans. 2*, 83 (1987) 1445-1452

.. [#Kline2006] S R Kline, *J Appl. Cryst.*, 39 (2006) 895

.. [#Onsager1949] L. Onsager, *Ann. New York Acad. Sci.*, 51 (1949) 627-659

Authorship and Verification
----------------------------

* **Author:** NIST IGOR/DANSE **Date:** pre 2010
* **Last Modified by:** Paul Kienzle **Date:** Aug 8, 2016
* **Last Reviewed by:** Richard Heenan **Date:** March 18, 2016
"""

import numpy as np
from numpy import pi, inf, sin, cos

name = "core_shell_cylinder"
title = "Right circular cylinder with a core-shell scattering length density profile."
description = """
P(q,alpha)= scale/Vs*f(q)^(2) + background,
      where: f(q)= 2(sld_core - solvant_sld)
        * Vc*sin[qLcos(alpha/2)]
        /[qLcos(alpha/2)]*J1(qRsin(alpha))
        /[qRsin(alpha)]+2(sld_shell-sld_solvent)
        *Vs*sin[q(L+T)cos(alpha/2)][[q(L+T)
        *cos(alpha/2)]*J1(q(R+T)sin(alpha))
        /q(R+T)sin(alpha)]

    alpha:is the angle between the axis of
        the cylinder and the q-vector
    Vs: the volume of the outer shell
    Vc: the volume of the core
    L: the length of the core
        sld_shell: the scattering length density of the shell
    sld_solvent: the scattering length density of the solvent
    background: the background
    T: the thickness
        R+T: is the outer radius
     L+2T: The total length of the outershell
    J1: the first order Bessel function
     theta: axis_theta of the cylinder
     phi: the axis_phi of the cylinder
"""
category = "shape:cylinder"

#             ["name", "units", default, [lower, upper], "type", "description"],
parameters = [["sld_core", "1e-6/Ang^2", 4, [-inf, inf], "sld",
               "Cylinder core scattering length density"],
              ["sld_shell", "1e-6/Ang^2", 4, [-inf, inf], "sld",
               "Cylinder shell scattering length density"],
              ["sld_solvent", "1e-6/Ang^2", 1, [-inf, inf], "sld",
               "Solvent scattering length density"],
              ["radius", "Ang", 20, [0, inf], "volume",
               "Cylinder core radius"],
              ["thickness", "Ang", 20, [0, inf], "volume",
               "Cylinder shell thickness"],
              ["length", "Ang", 400, [0, inf], "volume",
               "Cylinder length"],
              ["theta", "degrees", 60, [-360, 360], "orientation",
               "cylinder axis to beam angle"],
              ["phi", "degrees", 60, [-360, 360], "orientation",
               "rotation about beam"],
             ]

source = ["lib/polevl.c", "lib/sas_J1.c", "lib/gauss76.c", "core_shell_cylinder.c"]
have_Fq = True
radius_effective_modes = [
    "excluded volume", "equivalent volume sphere", "outer radius", "half outer length",
    "half min outer dimension", "half max outer dimension", "half outer diagonal",
    ]

def random():
    """Return a random parameter set for the model."""
    outer_radius = 10**np.random.uniform(1, 4.7)
    # Use a distribution with a preference for thin shell or thin core
    # Avoid core,shell radii < 1
    radius = np.random.beta(0.5, 0.5)*(outer_radius-2) + 1
    thickness = outer_radius - radius
    length = np.random.uniform(1, 4.7)
    pars = dict(
        radius=radius,
        thickness=thickness,
        length=length,
    )
    return pars

q = 0.1
# april 6 2017, rkh add unit tests, NOT compared with any other calc method, assume correct!
qx = q*cos(pi/6.0)
qy = q*sin(pi/6.0)
tests = [
    [{}, 0.075, 10.8552692237],
    [{}, (qx, qy), 0.444618752741],
]
del qx, qy  # not necessary to delete, but cleaner



r"""
Definition
----------

This model calculates an empirical functional form for SAS data characterized
by a broad scattering peak. Many SAS spectra are characterized by a broad peak
even though they are from amorphous soft materials. For example, soft systems
that show a SAS peak include copolymers, polyelectrolytes, multiphase systems,
layered structures, etc.

The d-spacing corresponding to the broad peak is a characteristic distance
between the scattering inhomogeneities (such as in lamellar, cylindrical, or
spherical morphologies, or for bicontinuous structures).

The scattering intensity $I(q)$ is calculated as

.. math:: I(q) = \frac{A}{q^n} + \frac{C}{1 + (|q - q_0|\xi)^m}^p + B

Here the peak position is related to the d-spacing as $q_0 = 2\pi / d_0$.

$A$ is the Porod law scale factor, $n$ the Porod exponent, $C$ is the
Lorentzian scale factor, $m$ the exponent of $q$, $\xi$ the screening length,
and $B$ the flat background. $p$ generalizes the model. With m = 2 and p = 1
the Lorentz model is obtained whereas for m = 2 and p = 2 the Broad-Peak model
is identical to the Debye-Anderson-Brumberger (dab) model.

For 2D data the scattering intensity is calculated in the same way as 1D,
where the $q$ vector is defined as

.. math:: q = \sqrt{q_x^2 + q_y^2}

References
----------

None.

Authorship and Verification
----------------------------

* **Author:** NIST IGOR/DANSE **Date:** pre 2010
* **Last Modified by:** Dirk Honecker **Date:** May 28, 2021
* **Last Reviewed by:** Richard Heenan **Date:** March 21, 2016
"""

import numpy as np
from numpy import inf, errstate

name = "broad_peak"
title = "Broad peak on top of a power law decay"
description = """\
      I(q) = scale_p/pow(q,exponent)+scale_l/
      pow(1.0 + pow(fabs(q-q_peak)*length_l,exponent_l), exponent_p )+ background

      List of default parameters:
      porod_scale = Porod term scaling
      porod_exp = Porod exponent
      peak_scale = Lorentzian term scaling
      correlation_length = Correlation length [A]
      peak_pos = peak location [1/A]
      width_exp = peak width exponent
      shape_exp = peak shape exponent
      background = Incoherent background"""
category = "shape-independent"

# pylint: disable=bad-whitespace, line-too-long
#             ["name", "units", default, [lower, upper], "type", "description"],
parameters = [["porod_scale",    "",  1.0e-05, [-inf, inf], "", "Power law scale factor"],
              ["porod_exp",      "",      3.0, [-inf, inf], "", "Exponent of power law"],
              ["peak_scale",  "",     10.0, [-inf, inf], "", "Scale factor for broad peak"],
              ["correlation_length", "Ang",  50.0, [-inf, inf], "", "screening length"],
              ["peak_pos",       "1/Ang", 0.1, [-inf, inf], "", "Peak position in q"],
              ["width_exp",    "",      2.0, [-inf, inf], "", "Exponent of peak width"],
              ["shape_exp",    "",      1.0, [-inf, inf], "", "Exponent of peak shape"],              
             ]
# pylint: enable=bad-whitespace, line-too-long

def Iq(q,
       porod_scale=1.0e-5,
       porod_exp=3.0,
       peak_scale=10.0,
       correlation_length=50.0,
       peak_pos=0.1,
       width_exp=2.0,
       shape_exp=1.0):
    """
    :param q:                    Input q-value
    :param porod_scale:          Power law scale factor
    :param porod_exp:            Exponent of power law
    :param peak_scale:           Scale factor for broad peak
    :param correlation_length:   Correlation length
    :param peak_pos:             Peak position in q
    :param width_exp:            Exponent of peak width
    :param shape_exp:            Exponent of peak shape    
    :return:                     Calculated intensity
    """
    z = abs(q - peak_pos) * correlation_length
    with errstate(divide='ignore'):
        inten = (porod_scale / q ** porod_exp
                 + peak_scale / (1 + z ** width_exp)) ** shape_exp
    return inten
Iq.vectorized = True  # Iq accepts an array of q values

def random():
    """Return a random parameter set for the model."""
    pars = dict(
        scale=1,
        porod_scale=10**np.random.uniform(-8, -5),
        porod_exp=np.random.uniform(1, 6),
        peak_scale=10**np.random.uniform(0.3, 6),
        correlation_length=10**np.random.uniform(0, 2),
        peak_pos=10**np.random.uniform(-3, -1),
        width_exp=np.random.uniform(1, 4),
        shape_exp=np.random.uniform(1, 2),      
    )
    pars['correlation_length'] /= pars['peak_pos']
    pars['peak_scale'] *= pars['porod_scale'] / pars['peak_pos']**pars['porod_exp']
    #pars['porod_scale'] = 0.
    return pars



#!/usr/bin/env python
r"""
Multiple scattering calculator

Calculate multiple scattering using 2D FFT convolution.

Usage:

.. code-block:: none

    python -m sasmodels.multiscat [options] model_name model_par=value ...

    Options include:
    -h, --help: show help and exit
    -n, --nq: the number of mesh points (dq = qmax*window/nq)
    -o, --outfile: save results to outfile.txt and outfile_powers.txt
    -p, --probability: the scattering probability (0.1)
    -q, --qmax: that max q that you care about (0.5)
    -r, --random: generate a random parameter set
    -s, --seed: generate a random parameter set with a given seed
    -w, --window: the extension window (q is calculated for qmax*window)
    -2, --2d: perform the calculation for an oriented pattern

Assume the probability of scattering is $p$. After each scattering event,
$1-p$ neutrons will leave the system and go to the detector, and the remaining
$p$ will scatter again.

Let the scattering probability for $n$ scattering event at $q$ be $f_n(q)$,
with

.. math:: f_1(q) = \frac{I_1(q)}{\int I_1(q) {\rm d}q}

for $I_1(q)$, the single scattering from the system. After two scattering
events, the scattering probability will be the convolution of the first
scattering and itself, or $f_2(q) = (f_1*f_1)(q)$.  After $n$ events it will be
$f_n(q) = (f_1 * \cdots * f_1)(q)$.  The total scattering is calculated
as the weighted sum of $f_k$, with weights following the Poisson distribution

.. math:: P(k; \lambda) = \frac{\lambda^k e^{-\lambda}}{k!}

for $\lambda$ determined by the total thickness divided by the mean
free path between scattering, giving

.. math:: I(q) = \sum_{k=0}^\infty P(k; \lambda) f_k(q)

The $k=0$ term is ignored since it involves no scattering.
We cut the series when cumulative probability is less than cutoff $C=99\%$,
which is $\max n$ such that

.. math:: \sum_{k=1}^n \frac{P(k; \lambda)}{1 - P(0; \lambda)} < C

Using the convolution theorem, where $F = \mathcal{F}(f)$ is the
Fourier transform,

.. math:: f * g = \mathcal{F}^{-1}\{\mathcal{F}\{f\} \cdot \mathcal{F}\{g\}\}

so

.. math:: f * \ldots * f = \mathcal{F}^{-1}\{ F^n \}

Since the Fourier transform is a linear operator, we can move the polynomial
expression for the convolution into the transform, giving

.. math::
    I(q) = \mathcal{F}^{-1}\left\{ \sum_{k=1}^{n} P(k; \lambda) F^k \right\}

In the dilute limit $\lambda \rightarrow 0$ only the $k=1$ term is active,
and so

.. math:: P(1; \lambda) = \lambda e^{-\lambda} = \int I_1(q) {\rm d}q

therefore we compute

.. math::
    I(q) =
    \mathcal{F}^{-1}\left\{
        \sum_{l=1}^{n} \frac{P(k; \lambda)}{P(1; \lambda))} F^k \right\}
    \, \int I_1(q) {\rm d}q
    = \mathcal{F}^{-1}\left\{
        \sum_{l=1}^{n} \frac{\lambda^{k-1}}{k!} F^k \right\}
    \, \int I_1(q) {\rm d}q

For speed we may use the fast fourier transform with a power of two.
The resulting $I(q)$ will be linearly spaced and likely heavily oversampled.
The usual pinhole or slit resolution calculation can performed from these
calculated values.

By default single precision OpenCL is used for the calculation.  Set the
environment variable *SAS_OPENCL=none* to use double precision numpy FFT
instead.  The OpenCL versions is about 10x faster on an elderly Mac with
Intel HD 4000 graphics.  The single precision numerical artifacts don't
seem to seriously impact overall accuracy, though they look pretty bad.
"""

from __future__ import print_function, division

import argparse
#import time

import numpy as np
from numpy import pi
from scipy.special import gamma

from sasmodels import core
from sasmodels import compare
from sasmodels.resolution import Resolution, bin_edges
from sasmodels.direct_model import call_kernel
import sasmodels.kernelcl

# TODO: select fast and accurate fft library
# clFFT: https://github.com/clMathLibraries/clFFT (AMD's OpenCL)
# - gpyfft: https://github.com/geggo/gpyfft (wraps clFFT)
# - arrayfire: https://github.com/arrayfire (wraps clFFT and much more; +cuda)
# pyFFT: https://github.com/fjarri-attic/pyfft (based on Apple's OpenCL; +cuda)
# - Reikna: https://github.com/fjarri/reikna (evolved from pyfft)
# genFFT: https://software.intel.com/en-us/articles/genFFT (Intel's OpenCL)
# VexCL: https://vexcl.readthedocs.io/en/latest/ (c++ library)
# TODO: switch to fftw when opencl is not available

try:
    import pyfft.cl
    import pyopencl.array as cl_array
    HAVE_OPENCL = sasmodels.kernelcl.use_opencl()
except ImportError:
    HAVE_OPENCL = False

# pylint: disable=unused-import,ungrouped-imports
try:
    from typing import Dict
    from sasmodels.modelinfo import ModelInfo
except ImportError:
    pass
# pylint: enable=unused-import,ungrouped-imports

PRECISION = np.dtype('f' if HAVE_OPENCL else 'd')  # 'f' or 'd'
USE_FAST = True  # OpenCL faster, less accurate math

class ICalculator(object):
    """
    Multiple scattering calculator
    """
    def fft(self, Iq):
        """
        Compute the forward FFT for an image, real -> complex.
        """
        raise NotImplementedError()

    def ifft(self, fourier_frame):
        """
        Compute the inverse FFT for an image, complex -> complex.
        """
        raise NotImplementedError()

    def multiple_scattering(self, Iq, p, coverage=0.99):
        r"""
        Compute multiple scattering for I(q) given scattering probability p.

        Given a probability p of scattering with the thickness, the expected
        number of scattering events, $\lambda = -\log(1 - p)$, giving a
        Poisson weighted sum of single, double, triple, etc. scattering patterns.
        The number of patterns used is based on coverage (default 99%).
        """
        raise NotImplementedError()

class NumpyCalculator(ICalculator):
    """
    Multiple scattering calculator using numpy fft.
    """
    def __init__(self, dims=None, dtype=PRECISION):
        self.dtype = dtype
        self.complex_dtype = np.dtype('F') if dtype == np.dtype('f') else np.dtype('D')

    def fft(self, Iq):
        #t0 = time.time()
        Iq = np.asarray(Iq, self.dtype)
        result = np.fft.fft2(Iq)
        #print("fft time", time.time()-t0)
        return result

    def ifft(self, fourier_frame):
        #t0 = time.time()
        fourier_frame = np.asarray(fourier_frame, self.complex_dtype)
        result = np.fft.ifft2(fourier_frame)
        #print("ifft time", time.time()-t0)
        return result

    def multiple_scattering(self, Iq, p, coverage=0.99):
        #t0 = time.time()
        coeffs = scattering_coeffs(p, coverage)
        poly = np.asarray(coeffs[::-1], dtype=self.dtype)
        scale = np.sum(Iq)
        frame = _forward_shift(Iq/scale, dtype=self.dtype)
        fourier_frame = np.fft.fft2(frame)
        convolved = fourier_frame * np.polyval(poly, fourier_frame)
        frame = np.fft.ifft2(convolved)
        result = scale * _inverse_shift(frame.real, dtype=self.dtype)
        #print("numpy multiscat time", time.time()-t0)
        return result

# polyval1(c, x) computes (...((c0 x + c1) x + c2) x ... + cn) x
# where c is an array of length *degree* and x is an array of
# complex values (type double2) of length *n*. 2-D arrays can of
# course be treated as 1-D arrays of length *nx* X *ny*.
# When compiling with sasmodels.kernelcl.compile_model the double precision
# types are converted to single precision as needed.  See the code in
# sasmodels.generate._convert_type for details.
POLYVAL1_KERNEL = """
kernel void polyval1(
    const int degree,
    global const double *coeff,
    const int n,
    global double2 *array)
{
    int index = get_global_id(0);
    if (index < n) {
        const double2 x = array[index];
        double2 total = coeff[0];
        for (int k=1; k < degree; k++) {
            total = fma(total, x, coeff[k]);
        }
        array[index] = total * x;
    }
}
"""

class OpenclCalculator(ICalculator):
    """
    Multiple scattering calculator using OpenCL via pyfft.
    """
    polyval1f = None
    polyval1d = None
    def __init__(self, dims, dtype=PRECISION):
        dtype = np.dtype(dtype)
        env = sasmodels.kernelcl.environment()
        context = env.context[dtype]
        if dtype == np.dtype('f'):
            if OpenclCalculator.polyval1f is None:
                program = sasmodels.kernelcl.compile_model(
                    context, POLYVAL1_KERNEL, dtype, fast=USE_FAST)
                # Assume context is always the same for a given dtype
                OpenclCalculator.polyval1f = program.polyval1
            self.dtype = dtype
            self.complex_dtype = np.dtype('F')
            self.polyval1 = OpenclCalculator.polyval1f
        else:
            if OpenclCalculator.polyval1d is None:
                program = sasmodels.kernelcl.compile_model(
                    context, POLYVAL1_KERNEL, dtype, fast=False)
                # Assume context is always the same for a given dtype
                OpenclCalculator.polyval1d = program.polyval1
            self.dtype = dtype
            self.complex_type = np.dtype('D')
            self.polyval1 = OpenclCalculator.polyval1d
        self.queue = env.queue[dtype]
        self.plan = pyfft.cl.Plan(dims, queue=self.queue)

    def fft(self, Iq):
        # forward transform
        #t0 = time.time()
        data = np.asarray(Iq, self.complex_dtype)
        gpu_data = cl_array.to_device(self.queue, data)
        self.plan.execute(gpu_data.data)
        result = gpu_data.get()
        #print("fft time", time.time()-t0)
        return result

    def ifft(self, fourier_frame):
        # inverse transform
        #t0 = time.time()
        data = np.asarray(fourier_frame, self.complex_dtype)
        gpu_data = cl_array.to_device(self.queue, data)
        self.plan.execute(gpu_data.data, inverse=True)
        result = gpu_data.get()
        #print("ifft time", time.time()-t0)
        return result

    def multiple_scattering(self, Iq, p, coverage=0.99):
        #t0 = time.time()
        coeffs = scattering_coeffs(p, coverage)
        scale = np.sum(Iq)
        poly = np.asarray(coeffs[::-1], self.dtype)
        frame = _forward_shift(Iq/scale, dtype=self.complex_dtype)
        gpu_data = cl_array.to_device(self.queue, frame)
        gpu_poly = cl_array.to_device(self.queue, poly)
        self.plan.execute(gpu_data.data)
        degree, data_size = poly.shape[0], frame.shape[0]*frame.shape[1]
        self.polyval1(
            self.queue, [data_size], None,
            np.int32(degree), gpu_poly.data, np.int32(data_size), gpu_data.data)
        self.plan.execute(gpu_data.data, inverse=True)
        frame = gpu_data.get()
        #result = scale * _inverse_shift(frame.real, dtype=self.dtype)
        result = scale * _inverse_shift(frame.real, dtype=self.dtype)
        #print("OpenCL multiscat time", time.time()-t0)
        return result

Calculator = OpenclCalculator if HAVE_OPENCL else NumpyCalculator

def scattering_powers(Iq, n, dtype='f', transform=None):
    r"""
    Calculate the scattering powers up to *n*.

    This includes *k=1* even though it should just be *Iq* itself

    The frames are unweighted; to weight scale by $\lambda^k e^{-\lambda}/k!$.
    """
    if transform is None:
        n_x, n_y = Iq.shape
        transform = Calculator(dims=(n_x*2, n_y*2), dtype=dtype)
    scale = np.sum(Iq)
    frame = _forward_shift(Iq/scale, dtype=dtype)
    F = transform.fft(frame)
    powers = [scale * _inverse_shift(transform.ifft(F**(k+1)).real, dtype=dtype)
              for k in range(n)]
    return powers

def scattering_coeffs(p, coverage=0.99):
    r"""
    Return the coefficients of the scattering powers for transmission
    probability *p*.  This is just the corresponding values for the
    Poisson distribution for $\lambda = -\ln(1-p)$ such that
    $\sum_{k = 0 \ldots n} P(k; \lambda)$ is larger than *coverage*.
    """
    L = -np.log(1-p)
    num_scatter = truncated_poisson_invcdf(coverage, L)
    coeffs = [L**k/gamma(k+2) for k in range(num_scatter)]
    return coeffs

def truncated_poisson_invcdf(coverage, L):
    r"""
    Return smallest k such that cdf(k; L) > coverage from the truncated Poisson
    probability excluding k=0
    """
    # pmf(k; L) = L**k * exp(-L) / (k! * (1-exp(-L))
    cdf = 0
    pmf = -np.exp(-L) / np.expm1(-L)
    k = 0
    while cdf < coverage:
        k += 1
        pmf *= L/k
        cdf += pmf
    return k

def _forward_shift(Iq, dtype=PRECISION):
    # Prepare padded array and forward transform
    nq = Iq.shape[0]
    half_nq = nq//2
    frame = np.zeros((2*nq, 2*nq), dtype=dtype)
    frame[:half_nq, :half_nq] = Iq[half_nq:, half_nq:]
    frame[-half_nq:, :half_nq] = Iq[:half_nq, half_nq:]
    frame[:half_nq, -half_nq:] = Iq[half_nq:, :half_nq]
    frame[-half_nq:, -half_nq:] = Iq[:half_nq, :half_nq]
    return frame

def _inverse_shift(frame, dtype=PRECISION):
    # Invert the transform and recover the transformed data
    nq = frame.shape[0]//2
    half_nq = nq//2
    Iq = np.empty((nq, nq), dtype=dtype)
    Iq[half_nq:, half_nq:] = frame[:half_nq, :half_nq]
    Iq[:half_nq, half_nq:] = frame[-half_nq:, :half_nq]
    Iq[half_nq:, :half_nq] = frame[:half_nq, -half_nq:]
    Iq[:half_nq, :half_nq] = frame[-half_nq:, -half_nq:]
    return Iq


class MultipleScattering(Resolution):
    r"""
    Compute multiple scattering using Fourier convolution.

    The fourier steps are determined by *qmax*, the maximum $q$ value
    desired, *nq* the number of $q$ steps and *window*, the amount
    of padding around the circular convolution.  The $q$ spacing
    will be $\Delta q = 2 q_\mathrm{max} w / n_q$.  If *nq* is not
    given it will use $n_q = 2^k$ such that $\Delta q < q_\mathrm{min}$.

    *probability* is related to the expected number of scattering
    events in the sample $\lambda$ as $p = 1 - e^{-\lambda}$.

    *coverage* determines how many scattering steps to consider.  The
    default is 0.99, which sets $n$ such that $1 \ldots n$ covers 99%
    of the Poisson probability mass function.

    *is2d* is True then 2D scattering is used, otherwise it accepts
    and returns 1D scattering.

    *resolution* is the resolution function to apply after multiple
    scattering.  If present, then the resolution $q$ vectors will provide
    default values for *qmin*, *qmax* and *nq*.
    """
    def __init__(self, qmin=None, qmax=None, nq=None, window=2,
                 probability=None, coverage=0.99,
                 is2d=False, resolution=None,
                 dtype=PRECISION):
        # Infer qmin, qmax from instrument resolution calculator, if present
        if resolution is not None:
            is2d = hasattr(resolution, 'qx_data')
            if is2d:
                # 2D data
                if qmax is None:
                    qx_calc, qy_calc = resolution.q_calc
                    qmax = np.sqrt(np.max(qx_calc**2 + qy_calc**2))
                if qmin is None and nq is None:
                    qx, qy = resolution.data.x_bins, resolution.data.y_bins
                    if qx and qy:
                        dx = (np.max(qx) - np.min(qx)) / len(qx)
                        dy = (np.max(qy) - np.min(qy)) / len(qy)
                    else:
                        qx, qy = resolution.data.qx_data, resolution.data.qy_data
                        steps = np.sqrt(len(qx))
                        dx = (np.max(qx) - np.min(qx)) / steps
                        dy = (np.max(qy) - np.min(qy)) / steps
                    qmin = min(dx, dy)
            else:
                # 1D data
                if qmax is None:
                    qmax = np.max(resolution.q_calc)
                if qmin is None and nq is None:
                    qmin = np.min(np.abs(resolution.q_calc))

        # estimate nq from qmin, qmax if not given explicitly
        q_range = qmax * window
        if nq is None:
            nq = 2**np.ceil(np.log2(q_range/qmin))
        nq = int(nq)
        # Compute available qmin based on nq
        qmin = 2*q_range / nq
        #print(nq)

        # remember input parameters
        self.qmax = qmax
        self.qmin = qmin
        self.nq = nq
        self.probability = 0. if probability is None else probability
        self.coverage = coverage
        self.is2d = is2d
        self.window = window
        self.resolution = resolution

        # Determine the q values to calculate
        q = np.linspace(-q_range, q_range, nq)
        qx, qy = np.meshgrid(q, q)
        if is2d:
            q_calc = (qx.flatten(), qy.flatten())
        else:
            # For 1-D patterns, compute q from the center to the corners and
            # interpolate from there into the individual pixels.  Given that
            # nq represents the points in [-qmax*windows, qmax*window],
            # then using 2*sqrt(2)*nq/2 will oversample the points in q by
            # a factor of two relative to the pixels.
            q_range_to_corner = np.sqrt(2.) * q_range
            nq_to_corner = 10*int(np.ceil(np.sqrt(2.) * nq))
            q_to_corner = np.linspace(0, q_range_to_corner, nq_to_corner+1)[1:]
            q_calc = (q_to_corner,)
            # Remember the q radii of the calculated points
            self._radius = np.sqrt(qx**2 + qy**2)
            #self._q = q_to_corner
        self._q_steps = q
        self.q_calc = q_calc

        # TODO: use cleaner data representation than that from sasview
        # Resolution function forwards underlying q data (for plotting, etc?)
        if is2d:
            if resolution is not None:
                # forward resolution function info to multiscattering
                self.qx_data = resolution.qx_data
                self.qy_data = resolution.qy_data
            else:
                # no underlying resolution function, but make it look like there is
                self.qx_data, self.qy_data = q_calc
        else:
            # 1-D radial profile is determined by the q values we need to
            # compute, either for the calculated q values for the resolution
            # function (if any) or for the raw q values desired
            self._q = np.linspace(qmin, qmax, int(nq//(2*window)))
            self._edges = bin_edges(self._q)
            self._norm, _ = np.histogram(self._radius, bins=self._edges)
            if resolution is not None:
                self.q = resolution.q
            else:
                # no underlying resolution function, but make it look like there is
                self.q = self._q

        # Prepare the multiple scattering calculator (either numpy or OpenCL)
        self.transform = Calculator((2*nq, 2*nq), dtype=dtype)

        # Iq and Iqxy will be set during apply
        self.Iq = None # type: np.ndarray
        self.Iqxy = None # type: np.ndarray

        # Label probability as a fittable parameter, and give its external name
        # Note that the external name must be a valid python identifier, since
        # is will be set as an experiment attribute.
        self.fittable = {'probability': 'scattering_probability'}

    def apply(self, theory):
        if self.is2d:
            Iq_calc = theory
        else:
            Iq_calc = np.interp(self._radius, self.q_calc[0], theory)
        Iq_calc = Iq_calc.reshape(self.nq, self.nq)

        # CRUFT: don't need probability as a function anymore
        probability = self.probability() if callable(self.probability) else self.probability
        coverage = self.coverage
        #t0 = time.time()
        Iqxy = self.transform.multiple_scattering(Iq_calc, probability, coverage)
        #print("multiple scattering calc time", time.time()-t0)

        # remember the intermediate result in case we want to see it later
        self.Iqxy = Iqxy

        if self.is2d:
            if self.resolution is not None:
                Iqxy = self.resolution.apply(Iqxy)
            return Iqxy
        else:
            # remember the intermediate result in case we want to see it later
            Iq = self.radial_profile(Iqxy)
            self.Iq = Iq
            if self.resolution is not None:
                q = self._q
                Iq_res = np.interp(np.abs(self.resolution.q_calc), q, self.Iq)
                _ = """
                k = 6
                print("q theory", q[:k])
                print("Iq theory", theory[:k])
                print("interp NaN?", np.any(np.isnan(Iq_calc)))
                print("convolved NaN?", np.any(np.isnan(Iqxy)))
                print("Iq intgrated", self.Iq[:k])
                print("radius", self._radius[self.nq/2,self.nq/2:self.nq/2+k])
                print("q edges", self._edges[:k+1])
                print("Iq norm", self._norm[:k])
                print("res q", self.resolution.q_calc[:k])
                print("Iq res", Iq_res[:k])
                #print(Iq)
                #print(Iq_res)
                """
                Iq = self.resolution.apply(Iq_res)
            return Iq

    def radial_profile(self, Iqxy):
        """
        Compute that radial profile for the given Iqxy grid.  The grid should
        be defined as for
        """
        # circular average, no anti-aliasing
        Iq = np.histogram(self._radius, bins=self._edges, weights=Iqxy)[0]/self._norm
        return Iq

    def plot_and_save_powers(self, theory, result, plot=True,
                              outfile="", background=0.):
        import pylab
        probability, coverage = self.probability, self.coverage
        weights = scattering_coeffs(probability, coverage)

        # cribbed from MultipleScattering.apply
        if self.is2d:
            Iq_calc = theory
        else:
            Iq_calc = np.interp(self._radius, self.q_calc[0], theory)
        Iq_calc = Iq_calc.reshape(self.nq, self.nq)

        # Compute the scattering powers for 1, 2, ... n scattering events
        powers = scattering_powers(Iq_calc, len(weights))

        #plotxy(Iqxy); import pylab; pylab.figure()
        if self.is2d:
            if outfile:
                data = np.vstack([Ipower.flatten() for Ipower in powers]).T
                np.savetxt(outfile + "_powers.txt", data)
                data = np.vstack(Iq_calc).T
                np.savetxt(outfile + ".txt", data)
            if plot:
                plotxy((self._q_steps, self._q_steps), Iq_calc)
                pylab.title("single scattering F")
                for k, v in enumerate(powers[1:]):
                    pylab.figure()
                    plotxy((self._q_steps, self._q_steps), v+background)
                    pylab.title("multiple scattering F^%d" % (k+2))
                pylab.figure()
                plotxy((self._q_steps, self._q_steps), self.Iqxy+background)
                pylab.title("total scattering for p=%g" % probability)
                if self.resolution is not None:
                    pylab.figure()
                    plotxy((self._q_steps, self._q_steps), result)
                    pylab.title("total scattering with resolution")
        else:
            q = self._q
            Iq_powers = [self.radial_profile(Iqxy) for Iqxy in powers]
            if outfile:
                data = np.vstack([q, theory, self.Iq]).T
                np.savetxt(outfile + ".txt", data)
                # circular average, no anti-aliasing for individual powers
                data = np.vstack([q] + Iq_powers).T
                np.savetxt(outfile + "_powers.txt", data)
            if plot:
                # Plot 2D pattern for total scattering
                plotxy((self._q_steps, self._q_steps), self.Iqxy+background)
                pylab.title("total scattering for p=%g" % probability)
                pylab.figure()

                # Plot 1D pattern for partial scattering
                pylab.loglog(q, self.Iq+background,
                             label="total for p=%g" % probability)
                if self.resolution is not None:
                    pylab.loglog(q, result, label="total with dQ")
                #new_annulus = annular_average(self._radius, self.Iqxy, self._edges)
                #pylab.loglog(q, new_annulus+background, label="new total for p=%g"%probability)
                for n, (w, Ipower) in enumerate(zip(weights, Iq_powers)):
                    pylab.loglog(q, w*Ipower+background,
                                 label="scattering^%d" % (n+1))
                pylab.legend()
                pylab.title('total scattering for p=%g' % probability)
        pylab.show()

def annular_average(qxy, Iqxy, qbins):
    """
    Compute annular average of points in *Iqxy* at *qbins*.  The $q_x$, $q_y$
    coordinates for *Iqxy* are given in *qxy*.
    """
    qxy, Iqxy = qxy.flatten(), Iqxy.flatten()
    index = np.argsort(qxy)
    qxy, Iqxy = qxy[index], Iqxy[index]
    print(qxy.shape, Iqxy.shape, index.shape, qbins.shape)
    #values = rebin(np.vstack((0., qxy)), Iqxy, qbins)
    integral = np.cumsum(Iqxy)
    Io = np.diff(np.interp(qbins, qxy, integral, left=0.))
    # normalize by area of annulus
    # TODO: doesn't properly account for box.
    # Need to chop off chords that are greater than the box edges
    # (left, right, top and bottom), then add back in the corners
    # chopped off by both. https://en.wikipedia.org/wiki/Circular_segment
    norms = np.diff(pi*qbins**2)
    return Io/norms

def rebin(x, I, xo):
    """
    Rebin from edges *x*, bins I into edges *xo*.

    *x* and *xo* should be monotonically increasing.

    If *x* has duplicate values, then all corresponding values at I(x) will
    be effectively summed into the same bin.  If *xo* has duplicate values,
    the first bin will contain the entire contents and subsequent bins will
    contain zeros.
    """
    integral = np.cumsum(I)
    Io = np.diff(np.interp(xo, x[1:], integral, left=0.))
    return Io


def parse_pars(model, opts):
    # type: (ModelInfo, argparse.Namespace) -> Dict[str, float]
    """
    Parse par=val arguments from the command line.
    """

    seed = np.random.randint(1000000) if opts.random and opts.seed < 0 else opts.seed
    compare_opts = {
        'info': (model.info, model.info),
        'seed': seed,
        'mono': True,
        'magnetic': False,
        'values': opts.pars,
        #'show_pars': True,
        'show_pars': False,
        'is2d': opts.is2d,
    }
    # Note: sascomp allows comparison on a pair of models, so ignore the second.
    pars, _ = compare.parse_pars(compare_opts)
    return pars


def main():
    """Command line interface to multiple scattering calculator."""
    parser = argparse.ArgumentParser(
        description="Compute multiple scattering",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
    parser.add_argument('-p', '--probability', type=float, default=0.1,
                        help="scattering probability")
    parser.add_argument('-n', '--nq', type=int, default=1024,
                        help='number of mesh points')
    parser.add_argument('-q', '--qmax', type=float, default=0.5,
                        help='max q')
    parser.add_argument('-w', '--window', type=float, default=2.0,
                        help='q calc = q max * window')
    parser.add_argument('-2', '--2d', dest='is2d', action='store_true',
                        help='oriented sample')
    parser.add_argument('-s', '--seed', default=-1,
                        help='random pars with given seed')
    parser.add_argument('-r', '--random', action='store_true',
                        help='random pars with random seed')
    parser.add_argument('-o', '--outfile', type=str, default="",
                        help='save to outfile.txt and outfile_powers.txt')
    parser.add_argument('model', type=str,
                        help='sas model name such as cylinder')
    parser.add_argument('pars', type=str, nargs='*',
                        help='model parameters such as radius=30')
    opts = parser.parse_args()
    assert opts.nq%2 == 0, "require even # points"

    model = core.load_model(opts.model)
    pars = parse_pars(model, opts)
    res = MultipleScattering(qmax=opts.qmax, nq=opts.nq, window=opts.window,
                             probability=opts.probability, is2d=opts.is2d)
    kernel = model.make_kernel(res.q_calc)
    #print(pars)
    bg = pars.get('background', 0.0)
    pars['background'] = 0.0
    theory = call_kernel(kernel, pars)
    Iq = res.apply(theory) + bg
    res.plot_and_save_powers(theory, Iq, outfile=opts.outfile, background=bg)

def plotxy(q, Iq):
    """Plot q, Iq or (qx, qy), Iqxy."""
    import pylab
    # q is a tuple of (q,) or (qx, qy)
    if len(q) == 1:
        pylab.loglog(q[0], Iq)
    else:
        data = Iq.copy()
        data[Iq <= 0] = np.min(Iq[Iq > 0])/2
        pylab.imshow(np.log10(data))

if __name__ == "__main__":
    main()
